# -*- coding: utf-8 -*-
# Copyright (c) 2017-2020 Richard Hull and contributors
# See LICENSE.rst for details.

from time import perf_counter

from PIL import Image, ImageDraw

from . import mixin
from .threadpool import threadpool

pool = threadpool(4)


def calc_bounds(xy, entity):
    """For an entity with width and height attributes, determine the bounding
    box if were positioned at ``(x, y)``."""
    left, top = xy
    right, bottom = left + entity.width, top + entity.height
    return [left, top, right, bottom]


def range_overlap(a_min, a_max, b_min, b_max):
    """Neither range is completely greater than the other."""
    return (a_min < b_max) and (b_min < a_max)


class viewport(mixin.capabilities):
    """The viewport offers a positionable window into a larger resolution
    pseudo-display, that also supports the concept of hotspots (which act like
    live displays).

    :param device: The device to project the enlarged pseudo-display viewport onto.
    :param width: The number of horizontal pixels.
    :type width: int
    :param height: The number of vertical pixels.
    :type height: int
    :param mode: The supported color model, one of ``"1"``, ``"RGB"`` or
        ``"RGBA"`` only.
    :type mode: str
    :param dither: By default, any color (other than black) will be `generally`
        treated as white when displayed on monochrome devices. However, this behaviour
        can be changed by adding ``dither=True`` and the image will be converted from RGB
        space into a 1-bit monochrome image where dithering is employed to differentiate
        colors at the expense of resolution.
    :type dither: bool
    """

    def __init__(self, device, width, height, mode=None, dither=False):
        self.capabilities(width, height, rotate=0, mode=mode or device.mode)
        if hasattr(device, "segment_mapper"):
            self.segment_mapper = device.segment_mapper
        self._device = device
        self._backing_image = Image.new(self.mode, self.size)
        self._position = (0, 0)
        self._hotspots = []
        self._dither = dither

    def display(self, image):
        assert image.mode == self.mode
        assert image.size == self.size

        self._backing_image.paste(image)
        self.refresh()

    def set_position(self, xy):
        self._position = xy
        self.refresh()

    def add_hotspot(self, hotspot, xy):
        """Add the hotspot at ``(x, y)``.

        The hotspot must fit inside the bounds of the virtual device. If
        it does not then an ``AssertError`` is raised.
        """
        (x, y) = xy
        assert 0 <= x <= self.width - hotspot.width
        assert 0 <= y <= self.height - hotspot.height

        # TODO: should it check to see whether hotspots overlap each other?
        # Is sensible to _allow_ them to overlap?
        self._hotspots.append((hotspot, xy))

    def remove_hotspot(self, hotspot, xy):
        """Remove the hotspot at ``(x, y)``: Any previously rendered image
        where the hotspot was placed is erased from the backing image, and will
        be "undrawn" the next time the virtual device is refreshed.

        If the specified hotspot is not found for ``(x, y)``, a
        ``ValueError`` is raised.
        """
        self._hotspots.remove((hotspot, xy))
        eraser = Image.new(self.mode, hotspot.size)
        self._backing_image.paste(eraser, xy)

    def is_overlapping_viewport(self, hotspot, xy):
        """Checks to see if the hotspot at position ``(x, y)`` is (at least
        partially) visible according to the position of the viewport."""
        l1, t1, r1, b1 = calc_bounds(xy, hotspot)
        l2, t2, r2, b2 = calc_bounds(self._position, self._device)
        return range_overlap(l1, r1, l2, r2) and range_overlap(t1, b1, t2, b2)

    def refresh(self):
        should_wait = False
        for hotspot, xy in self._hotspots:
            if hotspot.should_redraw() and self.is_overlapping_viewport(hotspot, xy):
                pool.add_task(hotspot.paste_into, self._backing_image, xy)
                should_wait = True

        if should_wait:
            pool.wait_completion()

        im = self._backing_image.crop(box=self._crop_box())
        if self._dither:
            im = im.convert(self._device.mode)

        self._device.display(im)

    def _crop_box(self):
        (left, top) = self._position
        right = left + self._device.width
        bottom = top + self._device.height

        assert 0 <= left <= right <= self.width
        assert 0 <= top <= bottom <= self.height

        return (left, top, right, bottom)


class hotspot(mixin.capabilities):
    """A hotspot (`a place of more than usual interest, activity, or
    popularity`)

    is a live display which may be added to a virtual viewport - if the hotspot
    and the viewport are overlapping, then the :func:`update` method will be
    automatically invoked when the viewport is being refreshed or its position
    moved (such that an overlap occurs).

    You would either:

        * create a ``hotspot`` instance, suppling a render function (taking an
          :py:mod:`PIL.ImageDraw` object, ``width`` & ``height`` dimensions.
          The render function should draw within a bounding box of ``(0, 0,
          width, height)``, and render a full frame.

        * sub-class ``hotspot`` and override the :func:`should_redraw` and
          :func:`update` methods. This might be more useful for slow-changing
          values where it is not necessary to update every refresh cycle, or
          your implementation is stateful.
    """

    def __init__(self, width, height, draw_fn=None):
        self.capabilities(width, height, rotate=0)  # TODO: set mode?
        self._fn = draw_fn

    def paste_into(self, image, xy):
        im = Image.new(image.mode, self.size)
        draw = ImageDraw.Draw(im)
        self.update(draw)
        image.paste(im, xy)
        del draw
        del im

    def should_redraw(self):
        """Override this method to return true or false on some condition
        (possibly on last updated member variable) so that for slow changing
        hotspots they are not updated too frequently."""
        return True

    def update(self, draw):
        if self._fn:
            self._fn(draw, self.width, self.height)


class snapshot(hotspot):
    """A snapshot is a `type of` hotspot, but only updates once in a given
    interval, usually much less frequently than the viewport requests refresh
    updates."""

    def __init__(self, width, height, draw_fn=None, interval=1.0):
        assert interval > 0
        super(snapshot, self).__init__(width, height, draw_fn)
        self.interval = interval
        self.last_updated = -interval

    def should_redraw(self):
        """Only requests a redraw after ``interval`` seconds have elapsed."""
        return perf_counter() - self.last_updated > self.interval

    def paste_into(self, image, xy):
        super(snapshot, self).paste_into(image, xy)
        self.last_updated = perf_counter()



from enum import IntEnum


class BrakingType(IntEnum):
    """Braking types."""

    COAST = 0
    BRAKE = 1


class ForwardDirection(IntEnum):
    """Forward directions."""

    CLOCKWISE = 1
    COUNTER_CLOCKWISE = -1


class Direction(IntEnum):
    """Directions."""

    FORWARD = 1
    BACK = -1



import logging
from math import pow

from pitop.common.i2c_device import I2CDevice

logger = logging.getLogger(__name__)

_i2c_device_name = "/dev/i2c-1"
_device_addr = 0x24
_register_addr = 0

_speaker_bit = 0
_mcu_bit = 1
_eeprom_bit = 2
_16khz_bit = 3

#######################
# INTERNAL OPERATIONS #
#######################


def __get_addr_for_bit(bit):
    if bit in [0, 1, 2, 3]:
        logger.debug("bit:  " + str(bit))
        addr = int(pow(2, bit))
        logger.debug("addr: " + str(addr))
        return addr
    else:
        logger.warning("Internal ERROR: invalid bit; cannot get address")
        return -1


def __get_bit_string(value):
    """INTERNAL.

    Get string representation of an int in binary
    """

    return "{0:b}".format(value).zfill(8)


def __update_device_state_bit(bit, value):
    """INTERNAL.

    Set a particular device state bit to enable or disable a particular
    function
    """

    # Bits:  0x0000
    # Index:   3210

    if bit not in [0, 1, 2, 3]:
        logger.warning("Error: Not a valid state bit")
        return False

    try:
        current_state = __read_device_state()
        logger.debug("Current device state: " + __get_bit_string(current_state))

    except Exception:
        logger.warning("Error: There was a problem getting the current device state")
        return False

    # Get the bit mask for the new state
    new_state = __get_addr_for_bit(bit)

    if value == 0:
        new_state = ~new_state

    # Check if there is anything to do
    if (value == 1 and (new_state & current_state) != 0) or (
        value == 0 and (~new_state & ~current_state) != 0
    ):
        logger.debug("Warning: Mode already set, nothing to send")
        return True

    if value == 0:
        new_state = new_state & current_state
    else:
        new_state = new_state | current_state

    # Combine the old with the new and send
    return __write_device_state(new_state)


def __verify_device_state(expected_state):
    """INTERNAL.

    Verify that that current device state matches that expected
    """

    current_state = __read_device_state()

    if expected_state == current_state:
        return True

    else:
        logger.warning(
            "Error: Device write verification failed. Expected: "
            + __get_bit_string(expected_state)
            + " Received: "
            + __get_bit_string(current_state)
        )
        return False


def __write_device_state(state):
    """INTERNAL.

    Send the state bits across the I2C bus to device
    """

    try:
        logger.debug("Connecting to device...")

        state_to_send = 0x0F & state

        logger.debug("Writing new state:    " + __get_bit_string(state_to_send))

        i2c_device = I2CDevice(_i2c_device_name, _device_addr)
        i2c_device.connect()
        i2c_device.write_byte(_register_addr, state_to_send)
        i2c_device.disconnect()

        result = __verify_device_state(state_to_send)

        if result is True:
            logger.debug("OK")
        else:
            logger.warning("Error: New state could not be verified")

        return result

    except Exception:
        logger.warning("Error: There was a problem writing to the device")
        return False


def __read_device_state():
    """INTERNAL.

    Read from the I2C device to get the current state of the pulse.
    Caller should handle exceptions
    """

    try:
        logger.debug("Connecting to device...")
        i2c_device = I2CDevice(_i2c_device_name, _device_addr)
        i2c_device.connect()
        current_state = i2c_device.read_unsigned_byte(_register_addr) & 0x0F
        i2c_device.disconnect()

        return int(current_state)

    except Exception:
        logger.warning("Error: There was a problem reading from the device")
        # Best to re-raise as we can't recover from this
        raise


def __reset_device_state(enable):
    """Reset the device state bits to the default enabled or disabled state."""

    clean_enable_state = __get_addr_for_bit(_eeprom_bit)
    clean_disable_state = __get_addr_for_bit(_speaker_bit) | __get_addr_for_bit(
        _mcu_bit
    )

    state_to_send = clean_enable_state if enable else clean_disable_state
    return __write_device_state(state_to_send)


#######################
# EXTERNAL OPERATIONS #
#######################


def reset_device_state(enable):
    """reset_device_state: Deprecated"""
    logger.info(
        "'reset_device_state' function has been deprecated, and can likely be removed. "
        "If you experience problems, please see documentation for instructions."
    )
    return False


def enable_device():
    logger.info(
        "'enable_device' function has been moved to pi-topd, and is handled automatically."
    )
    return False


def disable_device():
    logger.info(
        "'disable_device' function has been moved to pi-topd, and is handled automatically."
    )
    return False


def set_microphone_sample_rate_to_16khz():
    """Set the appropriate I2C bits to enable 16,000Hz recording on the
    microphone."""

    return __update_device_state_bit(_16khz_bit, 1)


def set_microphone_sample_rate_to_22khz():
    """Set the appropriate I2C bits to enable 22,050Hz recording on the
    microphone."""

    return __update_device_state_bit(_16khz_bit, 0)


# GET STATE


def speaker_enabled():
    """Get whether the speaker is enabled."""

    return (__read_device_state() & __get_addr_for_bit(_speaker_bit)) == 0


def mcu_enabled():
    """Get whether the onboard MCU is enabled."""

    return (__read_device_state() & __get_addr_for_bit(_mcu_bit)) == 0


def eeprom_enabled():
    """Get whether the eeprom is enabled."""

    return (__read_device_state() & __get_addr_for_bit(_eeprom_bit)) != 0


def microphone_sample_rate_is_16khz():
    """Get whether the microphone is set to record at a sample rate of
    16,000Hz."""

    return (__read_device_state() & __get_addr_for_bit(_16khz_bit)) != 0


def microphone_sample_rate_is_22khz():
    """Get whether the microphone is set to record at a sample rate of
    22,050Hz."""

    return (__read_device_state() & __get_addr_for_bit(_16khz_bit)) == 0



import logging
from fcntl import LOCK_EX, LOCK_NB, LOCK_UN, flock
from os import chmod
from os.path import exists
from stat import S_IWGRP, S_IWOTH, S_IWUSR
from threading import Lock

logger = logging.getLogger(__name__)


class PTLock(object):
    __lock_file_handle = None
    __locked_by_self = False

    def __init__(self, id):
        self.path = f"/tmp/.com.pi-top.sdk.{id}.lock"

        self._thread_lock = Lock()

        lock_file_already_existed = exists(self.path)
        self.__lock_file_handle = open(self.path, "w")
        if not lock_file_already_existed:
            chmod(self.path, S_IWUSR | S_IWGRP | S_IWOTH)

        logger.debug("Creating PTLock with path: {}".format(self.path))

    def acquire(self) -> bool:
        """Block until lock can be acquired."""
        if self._thread_lock.locked():
            logger.debug(
                "Attempting to acquire thread lock, which is currently already acquired."
            )

        if self.is_locked():
            logger.debug(
                f"Attempting to acquire lock file ({self.path}), which is currently already globally acquired."
            )

        logger.debug("Acquiring thread lock")
        self._thread_lock.acquire()
        logger.debug("Acquiring lock file at {}".format(self.path))
        flock(self.__lock_file_handle, LOCK_EX)
        self.__locked_by_self = True

    def release(self) -> bool:
        """Attempt to release lock."""
        if not self._thread_lock.locked():
            logger.debug(
                "Attempting to release thread lock, which is currently already acquired."
            )

        if not self.is_locked():
            logger.debug(
                f"Attempting to release lock file ({self.path}), which is currently already globally acquired."
            )

        logger.debug("Releasing thread lock")
        self._thread_lock.release()
        logger.debug("Releasing lock file at {}".format(self.path))
        flock(self.__lock_file_handle, LOCK_UN)
        self.__locked_by_self = False

    def is_locked(self):
        if self.__locked_by_self:
            return self.__locked_by_self

        lock_status = False
        try:
            flock(self.__lock_file_handle, LOCK_EX | LOCK_NB)
            flock(self.__lock_file_handle, LOCK_UN)
        except BlockingIOError:
            lock_status = True

        logger.debug(
            "Lock file at {} is {}locked".format(
                self.path, "not " if not lock_status else ""
            )
        )
        return lock_status

    __enter__ = acquire

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()

    def __del__(self):
        if self.__lock_file_handle:
            self.__lock_file_handle.close()



from enum import Enum
from threading import Lock

from pitop.pma.common import type_check

from .capture_actions import CaptureActions


class FrameHandler:
    """Class that processes actions performed on a camera frame.

    :data:`CaptureActions` must be registered, alongside an object using the :data:`register_action` method.
    These actions will be run whenever the :data:`process` method is called.
    """

    def __init__(self) -> None:
        self.__capture_actions = {}
        self.__frame = None

        self.__process_lock = Lock()
        self.__frame_lock = Lock()

    @property
    def frame(self):
        with self.__frame_lock:
            f = self.__frame
        return f

    @frame.setter
    def frame(self, frame):
        with self.__frame_lock:
            self.__frame = frame

    def process(self) -> None:
        """Executes all the actions registered in the FrameHandler object."""
        with self.__process_lock:
            frame = self.frame

            capture_actions = self.__capture_actions
            actions_to_remove = [CaptureActions.CAPTURE_SINGLE_FRAME]

            for action_name, action_objects in capture_actions.items():
                try:
                    action_objects.process(frame)
                except Exception as e:
                    print(f"Error processing {action_name}: {e}.")
                    actions_to_remove.append(action_name)

            for action_name in actions_to_remove:
                if self.is_running_action(action_name):
                    self.__capture_actions.pop(action_name)

    @type_check
    def register_action(self, action: CaptureActions, args_dict: dict) -> None:
        """Registers an action to be processed when running the :data:`process`
        method Based on the action, an action object instance will be created,
        using :data:`args_dict` as arguments.

        :param CaptureActions action: action to be registered
        :param dict args_dict: dictionary with arguments used to create an action object instance
        """

        if action in self.__capture_actions:
            print("Already registered this action.")
            return
        if "self" in args_dict:
            args_dict.pop("self")

        action_object = action.value(**args_dict)
        self.__capture_actions[action] = action_object

    @type_check
    def remove_action(self, action: CaptureActions) -> None:
        """Unregisters an action from the internal action register, to stop
        being processed when running :data:`process`

        :param CaptureActions action: type of action being removed
        """
        with self.__process_lock:
            if action in self.__capture_actions:
                action_object = self.__capture_actions.pop(action)
                action_object.stop()

    def current_actions(self) -> list:
        """Returns a list with the currently registered actions.

        :return: list
        """
        return [
            action.name if isinstance(action, Enum) else action
            for action in self.__capture_actions.keys()
        ]

    @type_check
    def is_running_action(self, action: CaptureActions) -> bool:
        """Checks if the given :data:`action` is in the internal action
        register.

        :param CaptureActions action: type of action to check
        :return: bool, True if the action is being processed, False otherwise
        """
        return action in self.__capture_actions



from pitop.miniscreen.oled.core.contrib.luma.core.const import common


class sh1106(common):
    CHARGEPUMP = 0x8D
    COLUMNADDR = 0x21
    COMSCANDEC = 0xC8
    COMSCANINC = 0xC0
    EXTERNALVCC = 0x1
    MEMORYMODE = 0x20
    PAGEADDR = 0x22
    SETCOMPINS = 0xDA
    SETDISPLAYCLOCKDIV = 0xD5
    SETDISPLAYOFFSET = 0xD3
    SETHIGHCOLUMN = 0x10
    SETLOWCOLUMN = 0x00
    SETPRECHARGE = 0xD9
    SETSEGMENTREMAP = 0xA1
    SETSTARTLINE = 0x40
    SETVCOMDETECT = 0xDB
    SWITCHCAPVCC = 0x2



import atexit
from dataclasses import dataclass

from pitop.core.mixins import Recreatable, Stateful
from pitop.pma.servo_controller import ServoController, ServoHardwareSpecs


@dataclass
class ServoMotorSetting:
    angle: float = 0.0
    speed: float = 0.0


class ServoMotor(Stateful, Recreatable):
    """Represents a pi-top servo motor component.

    Note that pi-top servo motors use an open-loop control system. As such, the output of the device (e.g.
    the angle and speed of the servo horn) cannot be measured directly. This means that you can set a target
    angle or speed for the servo, but you cannot read the current angle or speed.

    :type port_name: str
    :param port_name:
        The ID for the port to which this component is connected.

    :type zero_point: int
    :param zero_point:
        A user-defined offset from 'true' zero.
    """

    __HARDWARE_MIN_ANGLE = -ServoHardwareSpecs.ANGLE_RANGE / 2
    __HARDWARE_MAX_ANGLE = ServoHardwareSpecs.ANGLE_RANGE / 2
    __DEFAULT_SPEED = 50.0

    def __init__(self, port_name, zero_point=0, name="servo"):
        self._pma_port = port_name
        self.name = name

        self.__controller = ServoController(self._pma_port)

        self.__target_angle = 0.0
        self.__target_speed = self.__DEFAULT_SPEED

        self.__min_angle = self.__HARDWARE_MIN_ANGLE
        self.__max_angle = self.__HARDWARE_MAX_ANGLE
        self.__has_set_angle = False
        self.__zero_point = zero_point

        atexit.register(self.__cleanup)

        Stateful.__init__(self)
        Recreatable.__init__(
            self,
            config_dict={
                "port_name": port_name,
                "name": name,
                "zero_point": lambda: self.zero_point,
            },
        )

    @property
    def own_state(self):
        return {
            "angle": self.current_angle,
            "speed": self.current_speed,
        }

    def __cleanup(self):
        if self.__has_set_angle and self.current_speed != 0.0:
            self.stop()

    def stop(self):
        """Stop servo at its current position.

        :return: None
        """
        self.target_angle = self.current_angle

    @property
    def zero_point(self):
        """Represents the servo motor angle that the library treats as 'zero'.
        This value can be anywhere in the range of -90 to +90.

        For example, if the zero_point were set to be -30, then the valid range
        of values for setting the angle would be -60 to +120.

        .. warning::
            Setting a zero point out of the range of -90 to 90 will cause the method
            to raise an exception.
        """

        return self.__zero_point

    @zero_point.setter
    def zero_point(self, zero_position):
        if not (
            self.__HARDWARE_MIN_ANGLE <= zero_position <= self.__HARDWARE_MAX_ANGLE
        ):
            raise ValueError(
                f"Value must be from {self.__HARDWARE_MIN_ANGLE} to {self.__HARDWARE_MAX_ANGLE} degrees "
                f"(inclusive)"
            )

        self.__zero_point = zero_position
        self.__min_angle = self.__HARDWARE_MIN_ANGLE - self.__zero_point
        self.__max_angle = self.__HARDWARE_MAX_ANGLE - self.__zero_point

    @property
    def angle_range(self):
        """Returns a tuple with minimum and maximum possible angles where the
        servo horn can be moved to.

        If :class:`zero_point` is set to 0 (default), the angle range
        will be (-90, 90).
        """

        return self.__min_angle, self.__max_angle

    @property
    def setting(self):
        """Returns the current state of the servo motor, giving current angle
        and current speed.

        :return: :class:'ServoMotorSetting` object that has angle and speed attributes.
        """
        if not self.__has_set_angle:
            return None, None

        angle, speed = self.__controller.get_current_angle_and_speed()
        current_state = ServoMotorSetting()
        current_state.angle = angle - self.zero_point
        current_state.speed = speed
        return current_state

    @setting.setter
    def setting(self, target_state: ServoMotorSetting):
        """Sets the target state of the servo horn, relative to the zero
        position.

           .. warning::
             Using an :data:`target_state.angle` out of the valid angle range will cause the method to raise an
             exception. To determine the valid angle range, use :meth:`ServoMotor.get_angle_range`.

           .. warning::
             Using a :data:`target_state.speed` out of the valid speed range will cause the method to raise an exception.

        :type target_state: :class:`ServoMotorSetting`
        :param target_state:
            Set the target servo state using the :class:`ServoMotorSetting` class, both :meth:`ServoMotorSetting.speed` and
            :meth:`ServoMotorSetting.angle` must be passed. Example usage:

            .. code-block:: python
                from pitop import ServoMotor, ServoMotorSetting
                servo = ServoMotor()
                target_setting = ServoMotorSetting()
                target_setting.angle = 45
                target_setting.speed = 20
                servo.setting = target_setting
        """
        self.__controller.set_target_angle(
            target_state.angle + self.__zero_point, target_state.speed
        )
        self.__has_set_angle = True

    @property
    def current_angle(self):
        """Returns the current angle that the servo motor is at.

           .. note::
             If you need synchronized angle and speed values, use :meth:`ServoMotor.state` instead, this will return both
             current angle and current speed at the same time.

        :return: float value of the current angle of the servo motor in degrees.
        """
        if not self.__has_set_angle:
            raise RuntimeError(
                "Current angle is unknown. "
                "Please set a servo angle first to initialise the servo to an angle."
            )
        angle, _ = self.__controller.get_current_angle_and_speed()
        return angle - self.zero_point

    @property
    def current_speed(self):
        """Returns the current speed the servo motor is at.

           .. note::
             If you need synchronized angle and speed values, use :meth:`ServoMotor.state` instead, this will return both
             current angle and current speed at the same time.

        :return: float value of the current speed of the servo motor in deg/s.
        """
        _, speed = self.__controller.get_current_angle_and_speed()
        return speed

    @property
    def smooth_acceleration(self):
        """Gets whether or not the servo is configured to use a linear
        acceleration profile to ramp speed at start and end of cycle.

        :return: boolean value of the acceleration mode
        """
        return self.__controller.get_acceleration_mode() == 1

    @smooth_acceleration.setter
    def smooth_acceleration(self, enabled: bool):
        """Sets whether or not the servo is configured to use a linear
        acceleration profile to ramp speed at start and end of cycle.

        :type enabled: bool
        :param enabled: acceleration mode state
        """
        self.__controller.set_acceleration_mode(int(enabled))

    @property
    def target_angle(self):
        """Returns the last target angle that has been set.

        :return: float value of the target angle of the servo motor in deg.
        """
        return self.__target_angle

    @target_angle.setter
    def target_angle(self, angle):
        """Set the target angle you want the servo motor to go to.

        :type angle: float
        :param angle: target servo motor angle.
        """
        if not (self.__min_angle <= angle <= self.__max_angle):
            raise ValueError(
                f"Angle value must be from {self.__min_angle} to {self.__max_angle} degrees (inclusive)"
            )
        self.__target_angle = angle
        self.__controller.set_target_angle(
            self.__target_angle + self.__zero_point, self.__target_speed
        )
        self.__has_set_angle = True

    @property
    def target_speed(self):
        """Returns the last target speed that has been set.

        :return: float value of the target speed of the servo motor in deg/s.
        """
        return self.__target_speed

    @target_speed.setter
    def target_speed(self, speed):
        """Sets the servo motor speed. The speed value must be a number from.

        -100.0 to 100.0 deg/s.

           .. warning::
             Using a :data:`speed` out of the valid speed range will cause the method to raise an exception.

        :type speed: int or float
        :param speed:
            The target speed at which to move the servo horn, from -100 to 100 deg/s.
        """
        if not (
            -ServoHardwareSpecs.SPEED_RANGE <= speed <= ServoHardwareSpecs.SPEED_RANGE
        ):
            raise ValueError(
                f"Speed value must be from {ServoHardwareSpecs.SPEED_RANGE} to {ServoHardwareSpecs.SPEED_RANGE} deg/s (inclusive)"
            )
        self.__target_speed = speed

    def sweep(self, speed=None):
        """Moves the servo horn from the current position to one of the servo
        motor limits (maximum/minimum possible angle), moving at the specified
        speed. The speed value must be a number from -100.0 to 100.0 deg/s.

        The sweep direction is given by the speed.

        Setting a :data:`speed` value higher than zero will move the horn to the maximum angle (90 degrees by default),
        while a value less than zero will move it to the minimum angle (-90 degress by default).

           .. warning::
             Using a :data:`speed` out of the valid speed range will cause the method to raise an exception.

        :type speed: int or float
        :param speed:
            The target speed at which to move the servo horn, from -100 to 100 deg/s.
        """
        speed = self.target_speed if speed is None else speed
        if not (
            -ServoHardwareSpecs.SPEED_RANGE <= speed <= ServoHardwareSpecs.SPEED_RANGE
        ):
            raise ValueError(
                f"Speed value must be from {ServoHardwareSpecs.SPEED_RANGE} to {ServoHardwareSpecs.SPEED_RANGE} deg/s (inclusive)"
            )

        angle = self.__min_angle if speed < 0 else self.__max_angle
        self.__controller.set_target_angle(angle + self.__zero_point, speed)



import os

HERE = os.path.abspath(os.path.dirname(__file__))

__version__ = "N/A"
try:
    with open(f"{HERE}/version.txt", "r+") as f:
        # Reading form a file
        __version__ = f.read().strip()

except Exception:
    pass



import atexit
import logging
import time
from abc import abstractmethod
from collections import deque
from sched import scheduler
from threading import Event, Lock, Thread

import numpy as np
from gpiozero import SmoothedInputDevice

from pitop.common.common_ids import FirmwareDeviceID
from pitop.common.firmware_device import (
    FirmwareDevice,
    PTInvalidFirmwareDeviceException,
)
from pitop.common.singleton import Singleton
from pitop.pma.common.utils import get_pin_for_port

from .common.ultrasonic_registers import (
    UltrasonicConfigSettings,
    UltrasonicRegisters,
    UltrasonicRegisterTypes,
)
from .plate_interface import PlateInterface

logger = logging.getLogger(__name__)


class UltrasonicSensorBase:
    _max_distance = None
    threshold = None

    @property
    def max_distance(self):
        return self._max_distance

    @max_distance.setter
    def max_distance(self, value):
        if value <= 0:
            raise ValueError("invalid maximum distance (must be positive)")
        t = self.threshold_distance
        self._max_distance = value
        self.threshold_distance = t

    @property
    def threshold_distance(self):
        return self.threshold * self.max_distance

    @threshold_distance.setter
    def threshold_distance(self, value):
        self.threshold = value / self.max_distance

    @property
    def distance(self):
        return self.value * self._max_distance

    @abstractmethod
    def in_range(self):
        raise NotImplementedError

    @abstractmethod
    def value(self):
        raise NotImplementedError

    @abstractmethod
    def pin(self):
        raise NotImplementedError

    @abstractmethod
    def close(self):
        raise NotImplementedError

    @abstractmethod
    def wait_for_active(self, timeout=None):
        raise NotImplementedError

    @abstractmethod
    def wait_for_inactive(self, timeout=None):
        raise NotImplementedError


class CompatibilityCheck(metaclass=Singleton):
    __MIN_FIRMWARE_MAJOR_VERSION = 22

    def __init__(self):
        self.check()

    def check(self):
        try:
            firmware_device = FirmwareDevice(FirmwareDeviceID.pt4_expansion_plate)
            if (
                firmware_device.get_fw_version_major()
                < self.__MIN_FIRMWARE_MAJOR_VERSION
            ):
                raise RuntimeError(
                    "Usage of the analog ports for the Ultrasonic Sensor requires an Expansion Plate with "
                    f"a minimum version version of V{self.__MIN_FIRMWARE_MAJOR_VERSION}. "
                    f"Please update your Expansion Plate firmware to continue."
                )

        except PTInvalidFirmwareDeviceException:
            raise RuntimeError(
                "Please use an Expansion Plate in order to use the analog ports for the Ultrasonic "
                "Sensor."
            )


class UltrasonicSensorMCU(UltrasonicSensorBase):
    __MIN_FIRMWARE_MAJOR_VERSION = 22

    def __init__(
        self, port_name, queue_len, max_distance, threshold_distance, partial, name
    ):
        self._pma_port = port_name
        self.name = name

        # Distance readings
        if max_distance <= 0:
            raise ValueError("invalid maximum distance (must be positive)")
        self._max_distance = max_distance
        self._filtered_distance = max_distance
        self.threshold = threshold_distance / max_distance
        self._data_read_dt = 0.1
        self.__queue_len = queue_len
        self.__data_queue = deque(maxlen=self.__queue_len)

        # MCU configuration
        self.__mcu_device = PlateInterface().get_device_mcu()
        CompatibilityCheck()
        self.__registers = UltrasonicRegisters[self._pma_port]
        self.__configure_mcu()

        # User-programmable callbacks
        self.when_activated = None
        self.when_deactivated = None

        # Data state
        self.__partial = partial
        self.__data_ready = False
        self.__active = True

        # Thread communication
        self.__new_reading_event = Event()
        self.__activated_event = Event()
        self.__deactivated_event = Event()

        if self.__partial:
            self.__data_ready = True
        else:
            self.__queue_check = Thread(target=self.__queue_filled_check, daemon=True)
            self.__queue_check.start()

        # Data read loop
        self._read_scheduler = Thread(target=self.__read_scheduler, daemon=True)
        self._read_scheduler.start()

        # Monitor for changes from active to inactive or vice versa
        self.__state_check = Thread(target=self.__state_monitor, daemon=True)
        self.__state_check.start()

        atexit.register(self.close)

    def __configure_mcu(self):
        self.__mcu_device.write_byte(
            self.__registers[UltrasonicRegisterTypes.CONFIG],
            UltrasonicConfigSettings[self._pma_port],
        )

    @property
    def value(self):
        while not self.__data_ready:
            time.sleep(self._data_read_dt)
        return min(1.0, self._filtered_distance / self._max_distance)

    @property
    def pin(self):
        print(
            "An Ultrasonic Sensor connected to an analog port is controlled directly by the MCU and does not have an"
            "associated Raspberry Pi pin. Returning None."
        )
        return None

    def close(self):
        self.__mcu_device.write_byte(
            self.__registers[UltrasonicRegisterTypes.CONFIG], 0x00
        )

    @property
    def in_range(self):
        return not self.__active

    def wait_for_active(self, timeout=None):
        self.__activated_event.wait(timeout=timeout)

    def wait_for_inactive(self, timeout=None):
        self.__deactivated_event.wait(timeout=timeout)

    def __read_scheduler(self):
        s = scheduler(time.time, time.sleep)
        s.enter(self._data_read_dt, 1, self.__read_loop, (s,))
        s.run()

    def __read_loop(self, s):
        self.__data_queue.append(self.__read_distance())
        self._filtered_distance = np.median(self.__data_queue)
        self.__new_reading_event.set()
        self.__new_reading_event.clear()
        s.enter(self._data_read_dt, 1, self.__read_loop, (s,))

    def __state_monitor(self):
        while True:
            self.__new_reading_event.wait()
            if self.__data_ready:
                self.__check_for_state_change()

    def __check_for_state_change(self):
        if self.__active and self.__inactive_criteria():
            self.__was_deactivated()
            return
        if not self.__active and self.__active_criteria():
            self.__was_activated()
            return

    def __active_criteria(self):
        return self._filtered_distance >= self.threshold_distance

    def __inactive_criteria(self):
        return not self.__active_criteria()

    def __was_activated(self):
        if callable(self.when_activated):
            self.when_activated()
        self.__active = True
        self.__activated_event.set()
        self.__activated_event.clear()

    def __was_deactivated(self):
        if callable(self.when_deactivated):
            self.when_deactivated()
        self.__active = False
        self.__deactivated_event.set()
        self.__deactivated_event.clear()

    def __queue_filled_check(self):
        while True:
            self.__new_reading_event.wait()
            if self.__queue_len == len(self.__data_queue):
                self.__data_ready = True
                break

    def __read_distance(self):
        distance = (
            self.__mcu_device.read_unsigned_word(
                register_address=self.__registers[UltrasonicRegisterTypes.DATA],
                little_endian=True,
            )
            / 100
        )
        if distance == 0:
            return self._max_distance
        return distance


# Modified version of gpiozero's DistanceSensor class that only uses 1 pin
#
# Note: all private member variables are semi-private to follow upstream gpiozero convention
# and to override inherited functions
class UltrasonicSensorRPI(SmoothedInputDevice, UltrasonicSensorBase):
    ECHO_LOCK = Lock()

    def __init__(
        self,
        port_name,
        queue_len=3,
        max_distance=3,
        threshold_distance=0.3,
        partial=True,
        name="ultrasonic",
    ):

        self._pma_port = port_name
        self.name = name

        SmoothedInputDevice.__init__(
            self,
            get_pin_for_port(self._pma_port),
            pull_up=False,
            queue_len=queue_len,
            sample_wait=0.1,
            partial=partial,
            ignore=frozenset({None}),
        )

        try:
            if max_distance <= 0:
                raise ValueError("invalid maximum distance (must be positive)")
            self._max_distance = max_distance
            self.threshold = threshold_distance / max_distance
            self.speed_of_sound = 343.26  # m/s

            self._echo = Event()
            self._echo_rise = None
            self._echo_fall = None
            self.pin.edges = "both"
            self.pin.bounce = None
            self.pin.when_changed = self._echo_changed
            self._queue.start()
        except Exception:
            self.close()
            raise

    def close(self):
        try:
            super(UltrasonicSensorRPI, self).close()
        except RuntimeError:
            logger.debug(
                f"Ultrasonic Sensor on port {self._pma_port} - "
                "there was an error in closing the port!"
            )

    @property
    def value(self):
        return super(UltrasonicSensorRPI, self).value

    @property
    def pin(self):
        return super(UltrasonicSensorRPI, self).pin

    def _echo_changed(self, ticks, level):
        if level:
            self._echo_rise = ticks
        else:
            self._echo_fall = ticks
            self._echo.set()

    def _read(self):
        # Wait up to 50ms for the echo pin to fall to low (the maximum echo
        # pulse is 35ms so this gives some leeway); if it doesn't something is
        # horribly wrong (most likely at the hardware level)
        if self.pin.state:
            if not self._echo.wait(0.05):
                logger.debug(
                    f"Ultrasonic Sensor on port {self._pma_port} - "
                    "no echo received, not using value"
                )
                return None
        self._echo.clear()
        self._echo_fall = None
        self._echo_rise = None
        # Obtain the class-level ECHO_LOCK to ensure multiple distance sensors
        # don't listen for each other's "pings"
        with UltrasonicSensorRPI.ECHO_LOCK:
            # Wait up to 200ms for the echo pin to rise and fall
            if self._echo.wait(0.2):
                if self._echo_fall is not None and self._echo_rise is not None:
                    distance = (
                        self.pin_factory.ticks_diff(self._echo_fall, self._echo_rise)
                        * self.speed_of_sound
                        / 2.0
                    )
                    return round(min(1.0, distance / self._max_distance), 2)
                else:
                    # If we only saw the falling edge it means we missed
                    # the echo because it was too fast
                    return None
            else:
                # The echo pin never rose or fell - assume that distance is max
                logger.debug(
                    f"Ultrasonic Sensor on port {self._pma_port} - "
                    "no echo received, using max distance "
                )
                return 1.0

    @property
    def in_range(self):
        return not self.is_active



# coding=utf-8
# pynput
# Copyright (C) 2015-2018 Moses Palmér
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""The mouse implementation for *Windows*."""

# pylint: disable=C0111
# The documentation is extracted from the base classes

# pylint: disable=R0903
# We implement stubs

import ctypes
import enum
from ctypes import windll, wintypes

from pynput._util import NotifierMixin
from pynput._util.win32 import (
    INPUT,
    MOUSEINPUT,
    INPUT_union,
    ListenerMixin,
    SendInput,
    SystemHook,
)

from . import _base


class Button(enum.Enum):
    """The various buttons."""

    unknown = None
    left = (MOUSEINPUT.LEFTUP, MOUSEINPUT.LEFTDOWN)
    middle = (MOUSEINPUT.MIDDLEUP, MOUSEINPUT.MIDDLEDOWN)
    right = (MOUSEINPUT.RIGHTUP, MOUSEINPUT.RIGHTDOWN)


class Controller(NotifierMixin, _base.Controller):
    __GetCursorPos = windll.user32.GetCursorPos
    __SetCursorPos = windll.user32.SetCursorPos

    def __init__(self, *args, **kwargs):
        super(Controller, self).__init__(*args, **kwargs)

    def _position_get(self):
        point = wintypes.POINT()
        if self.__GetCursorPos(ctypes.byref(point)):
            return (point.x, point.y)
        else:
            return None

    def _position_set(self, pos):
        pos = int(pos[0]), int(pos[1])
        self.__SetCursorPos(*pos)
        self._emit("on_move", *pos)

    def _scroll(self, dx, dy):
        if dy:
            SendInput(
                1,
                ctypes.byref(
                    INPUT(
                        type=INPUT.MOUSE,
                        value=INPUT_union(
                            mi=MOUSEINPUT(dwFlags=MOUSEINPUT.WHEEL, mouseData=int(dy))
                        ),
                    )
                ),
                ctypes.sizeof(INPUT),
            )

        if dx:
            SendInput(
                1,
                ctypes.byref(
                    INPUT(
                        type=INPUT.MOUSE,
                        value=INPUT_union(
                            mi=MOUSEINPUT(dwFlags=MOUSEINPUT.HWHEEL, mouseData=int(dx))
                        ),
                    )
                ),
                ctypes.sizeof(INPUT),
            )

        if dx or dy:
            px, py = self._position_get()
            self._emit("on_scroll", px, py, dx, dy)

    def _press(self, button):
        SendInput(
            1,
            ctypes.byref(
                INPUT(
                    type=INPUT.MOUSE,
                    value=INPUT_union(mi=MOUSEINPUT(dwFlags=button.value[1])),
                )
            ),
            ctypes.sizeof(INPUT),
        )

    def _release(self, button):
        SendInput(
            1,
            ctypes.byref(
                INPUT(
                    type=INPUT.MOUSE,
                    value=INPUT_union(mi=MOUSEINPUT(dwFlags=button.value[0])),
                )
            ),
            ctypes.sizeof(INPUT),
        )


@Controller._receiver
class Listener(ListenerMixin, _base.Listener):
    #: The Windows hook ID for low level mouse events, ``WH_MOUSE_LL``
    _EVENTS = 14

    WM_LBUTTONDOWN = 0x0201
    WM_LBUTTONUP = 0x0202
    WM_MBUTTONDOWN = 0x0207
    WM_MBUTTONUP = 0x0208
    WM_MOUSEMOVE = 0x0200
    WM_MOUSEWHEEL = 0x020A
    WM_MOUSEHWHEEL = 0x020E
    WM_RBUTTONDOWN = 0x0204
    WM_RBUTTONUP = 0x0205

    _WHEEL_DELTA = 120

    #: A mapping from messages to button events
    CLICK_BUTTONS = {
        WM_LBUTTONDOWN: (Button.left, True),
        WM_LBUTTONUP: (Button.left, False),
        WM_MBUTTONDOWN: (Button.middle, True),
        WM_MBUTTONUP: (Button.middle, False),
        WM_RBUTTONDOWN: (Button.right, True),
        WM_RBUTTONUP: (Button.right, False),
    }

    #: A mapping from messages to scroll vectors
    SCROLL_BUTTONS = {WM_MOUSEWHEEL: (0, 1), WM_MOUSEHWHEEL: (1, 0)}

    _HANDLED_EXCEPTIONS = (SystemHook.SuppressException,)

    class _MSLLHOOKSTRUCT(ctypes.Structure):
        """Contains information about a mouse event passed to a ``WH_MOUSE_LL``
        hook procedure, ``MouseProc``."""

        _fields_ = [
            ("pt", wintypes.POINT),
            ("mouseData", wintypes.DWORD),
            ("flags", wintypes.DWORD),
            ("time", wintypes.DWORD),
            ("dwExtraInfo", ctypes.c_void_p),
        ]

    #: A pointer to a :class:`_MSLLHOOKSTRUCT`
    _LPMSLLHOOKSTRUCT = ctypes.POINTER(_MSLLHOOKSTRUCT)

    def __init__(self, *args, **kwargs):
        super(Listener, self).__init__(*args, **kwargs)
        self._event_filter = self._options.get("event_filter", lambda msg, data: True)

    def _handle(self, code, msg, lpdata):
        if code != SystemHook.HC_ACTION:
            return

        data = ctypes.cast(lpdata, self._LPMSLLHOOKSTRUCT).contents

        # Suppress further propagation of the event if it is filtered
        if self._event_filter(msg, data) is False:
            return

        if msg == self.WM_MOUSEMOVE:
            self.on_move(data.pt.x, data.pt.y)

        elif msg in self.CLICK_BUTTONS:
            button, pressed = self.CLICK_BUTTONS[msg]
            self.on_click(data.pt.x, data.pt.y, button, pressed)

        elif msg in self.SCROLL_BUTTONS:
            mx, my = self.SCROLL_BUTTONS[msg]
            dd = wintypes.SHORT(data.mouseData >> 16).value // self._WHEEL_DELTA
            self.on_scroll(data.pt.x, data.pt.y, dd * mx, dd * my)



import logging
import signal
from binascii import hexlify, unhexlify
from os import close, path, remove, rename, stat
from struct import pack
from sys import exit
from tempfile import mkstemp
from threading import Thread
from time import sleep

import serial

from pitop.pulse import configuration

logger = logging.getLogger(__name__)

_bitrate = 8
_continue_writing = False
_recording_thread = False
_thread_running = False
_exiting = False
_temp_file_path = ""

#######################
# INTERNAL OPERATIONS #
#######################


def __signal_handler(signal, frame):
    """INTERNAL.

    Handles signals from the OS.
    """

    global _exiting

    if _exiting is False:
        _exiting = True

        if _thread_running is True:
            stop()

    logger.info("\nQuitting...")
    exit(0)


def __get_size(filename):
    """INTERNAL.

    Gets the size of a file.
    """

    file_stats = stat(filename)
    return file_stats.st_size


def __from_hex(value):
    """INTERNAL.

    Gets a bytearray from hex data.
    """

    return bytearray.fromhex(value)


def __space_separated_little_endian(integer_value, byte_len):
    """INTERNAL.

    Get an integer in format for WAV file header.
    """

    if byte_len <= 1:
        pack_type = "<B"
    elif byte_len <= 2:
        pack_type = "<H"
    elif byte_len <= 4:
        pack_type = "<I"
    elif byte_len <= 8:
        pack_type = "<Q"
    else:
        logger.info("Value cannot be represented in 8 bytes - exiting")
        exit()

    hex_string = pack(pack_type, integer_value)
    temp = hexlify(hex_string).decode()
    return " ".join([temp[i : i + 2] for i in range(0, len(temp), 2)])


def __init_header_information():
    """INTERNAL.

    Create a WAV file header.
    """

    RIFF = "52 49 46 46"
    WAVE = "57 41 56 45"
    fmt = "66 6d 74 20"
    DATA = "64 61 74 61"

    if configuration.microphone_sample_rate_is_22khz():
        capture_sample_rate = 22050
    else:
        capture_sample_rate = 16000

    # ChunkID
    header = __from_hex(RIFF)
    # ChunkSize - 4 bytes (to be changed depending on length of data...)
    header += __from_hex(__space_separated_little_endian(0, 4))
    # Format
    header += __from_hex(WAVE)
    # Subchunk1ID
    header += __from_hex(fmt)
    # Subchunk1Size (PCM = 16)
    header += __from_hex(__space_separated_little_endian(16, 4))
    # AudioFormat   (PCM = 1)
    header += __from_hex(__space_separated_little_endian(1, 2))
    header += __from_hex(__space_separated_little_endian(1, 2))  # NumChannels
    # SampleRate
    header += __from_hex(__space_separated_little_endian(capture_sample_rate, 4))
    # ByteRate (Same as SampleRate due to 1 channel, 1 byte per sample)
    header += __from_hex(__space_separated_little_endian(capture_sample_rate, 4))
    # BlockAlign - (no. of bytes per sample)
    header += __from_hex(__space_separated_little_endian(1, 2))
    # BitsPerSample
    header += __from_hex(__space_separated_little_endian(_bitrate, 2))
    # Subchunk2ID
    header += __from_hex(DATA)
    # Subchunk2Size - 4 bytes (to be changed depending on length of data...)
    header += __from_hex(__space_separated_little_endian(0, 4))

    return header


def __update_header_in_file(file, position, value):
    """INTERNAL.

    Update the WAV header
    """

    hex_value = __space_separated_little_endian(value, 4)
    data = unhexlify("".join(hex_value.split()))

    file.seek(position)
    file.write(data)


def __finalise_wav_file(file_path):
    """INTERNAL.

    Update the WAV file header with the size of the data.
    """

    size_of_data = __get_size(file_path) - 44

    if size_of_data <= 0:
        logger.info("Error: No data was recorded!")
        remove(file_path)
    else:
        with open(file_path, "rb+") as file:

            logger.debug("Updating header information...")

            __update_header_in_file(file, 4, size_of_data + 36)
            __update_header_in_file(file, 40, size_of_data)


def __thread_method():
    """INTERNAL.

    Thread method.
    """

    __record_audio()


def __record_audio():
    """INTERNAL.

    Open the serial port and capture audio data into a temp file.
    """

    global _temp_file_path

    temp_file_tuple = mkstemp()
    close(temp_file_tuple[0])
    _temp_file_path = temp_file_tuple[1]

    if path.exists("/dev/serial0"):

        logger.debug("Opening serial device...")

        serial_device = serial.Serial(
            port="/dev/serial0",
            timeout=1,
            baudrate=250000,
            parity=serial.PARITY_NONE,
            stopbits=serial.STOPBITS_ONE,
            bytesize=serial.EIGHTBITS,
        )
        serial_device_open = serial_device.isOpen()

        if serial_device_open is True:

            try:
                logger.debug("Start recording")

                with open(_temp_file_path, "wb") as file:

                    logger.debug("WRITING: initial header information")
                    file.write(__init_header_information())

                    if serial_device.inWaiting():
                        logger.debug("Flushing input and starting from scratch")
                        serial_device.flushInput()

                    logger.debug("WRITING: wave data")

                    while _continue_writing:
                        while not serial_device.inWaiting():
                            sleep(0.01)

                        audio_output = serial_device.read(serial_device.inWaiting())
                        bytes_to_write = bytearray()

                        for pcm_data_block in audio_output:

                            if _bitrate == 16:

                                pcm_data_int = 0
                                pcm_data_int = pcm_data_block
                                scaled_val = int((pcm_data_int * 32768) / 255)
                                bytes_to_write += __from_hex(
                                    __space_separated_little_endian(scaled_val, 2)
                                )

                            else:

                                pcm_data_int = pcm_data_block
                                bytes_to_write += __from_hex(
                                    __space_separated_little_endian(pcm_data_int, 1)
                                )

                        file.write(bytes_to_write)

                        sleep(0.1)

            finally:
                serial_device.close()

                __finalise_wav_file(_temp_file_path)

                logger.debug("Finished Recording.")

        else:
            logger.info("Error: Serial port failed to open")

    else:
        logger.info("Error: Could not find serial port - are you sure it's enabled?")


#######################
# EXTERNAL OPERATIONS #
#######################


def record():
    """Start recording on the pi-topPULSE microphone."""

    global _thread_running
    global _continue_writing
    global _recording_thread

    if not configuration.mcu_enabled():
        logger.info("Error: pi-topPULSE is not initialised.")
        exit()

    if _thread_running is False:
        _thread_running = True
        _continue_writing = True
        _recording_thread = Thread(group=None, target=__thread_method)
        _recording_thread.start()
    else:
        logger.info("Microphone is already recording!")


def is_recording():
    """Returns recording state of the pi-topPULSE microphone."""

    return _thread_running


def stop():
    """Stops recording audio."""

    global _thread_running
    global _continue_writing

    _continue_writing = False
    _recording_thread.join()
    _thread_running = False


def save(file_path, overwrite=False):
    """Saves recorded audio to a file."""

    global _temp_file_path

    if _thread_running is False:
        if _temp_file_path != "" and path.exists(_temp_file_path):
            if path.exists(file_path) is False or overwrite is True:

                if path.exists(file_path):
                    remove(file_path)

                rename(_temp_file_path, file_path)
                _temp_file_path = ""

            else:
                logger.info("File already exists")
        else:
            logger.info("No recorded audio data found")
    else:
        logger.info("Microphone is still recording!")


def set_sample_rate_to_16khz():
    """Set the appropriate I2C bits to enable 16,000Hz recording on the
    microphone."""

    configuration.set_microphone_sample_rate_to_16khz()


def set_sample_rate_to_22khz():
    """Set the appropriate I2C bits to enable 22,050Hz recording on the
    microphone."""

    configuration.set_microphone_sample_rate_to_22khz()


def set_bit_rate_to_unsigned_8():
    """Set bitrate to device default."""

    global _bitrate
    _bitrate = 8


def set_bit_rate_to_signed_16():
    """Set bitrate to double that of device default by scaling the signal."""

    global _bitrate
    _bitrate = 16


#######################
# INITIALISATION      #
#######################

_signal = signal.signal(signal.SIGINT, __signal_handler)



from argparse import ArgumentParser

from .cli_base import CliBaseClass, PitopCliException, PitopCliInvalidArgument


class DisplayCLI(CliBaseClass):
    parser_help = "communicate and control the device's display"
    cli_name = "display"

    def __init__(self, args) -> None:
        self.args = args
        self.validate_args()

    def run(self) -> int:
        try:
            self.perform_desired_action()
            return 0
        except Exception as e:
            print(f"Error on pitop-display.run: {e}")
            return 1

    def validate_args(self) -> None:
        if vars(self.args).get("display_subcommand") is None:
            raise PitopCliInvalidArgument
        # Handle invalid command line parameter combinations
        if self.args.display_subcommand != "brightness":
            return
        if self.args.brightness_value and (
            self.args.increment_brightness or self.args.decrement_brightness
        ):
            print(
                "Error on pitop-display.validate_args: Cannot increment/decrement at the same time as setting brightness value"
            )
            raise PitopCliException
        if self.args.increment_brightness and self.args.decrement_brightness:
            print(
                "Error on pitop-display.validate_args: Cannot increment and decrement brightness at the same time"
            )
            raise PitopCliException

    def perform_desired_action(self) -> None:
        from pitop.display import Display

        display = Display()

        if self.args.display_subcommand == "brightness":
            if self.args.increment_brightness:
                display.increment_brightness()
            elif self.args.decrement_brightness:
                display.decrement_brightness()
            elif (
                vars(self.args).get("brightness_value")
                or self.args.brightness_value is not None
            ):
                display.brightness = self.args.brightness_value
            else:
                print(display.brightness)

        elif self.args.display_subcommand == "backlight":
            if vars(self.args).get("backlight_value") is None:
                print(display.backlight)
            else:
                display.backlight = self.args.backlight_value in ("on", 1)

        elif self.args.display_subcommand == "blank_time":
            if vars(self.args).get("timeout_value") is None:
                print(display.blanking_timeout)
            else:
                display.blanking_timeout = self.args.timeout_value

    @classmethod
    def add_parser_arguments(cls, parser) -> None:
        subparser = parser.add_subparsers(
            title="pi-top display utility",
            description="Interface to communicate with pi-top device's onboard display",
            dest="display_subcommand",
        )

        # "pi-top display brightness"
        brightness_parser = subparser.add_parser(
            "brightness",
            help="Control onboard pi-top display brightness (Original pi-top, pi-topCEED, pi-top [3])",
        )
        cls.add_brightness_arguments(brightness_parser)

        # "pi-top display backlight"
        backlight_parser = subparser.add_parser(
            "backlight", help="Control onboard pi-top display backlight"
        )
        backlight_parser.add_argument(
            "backlight_value",
            help="Set the screen backlight state",
            choices=("on", "off", 0, 1),
            nargs="?",
        )

        # "pi-top display blank_time"
        timeout_parser = subparser.add_parser(
            "blank_time",
            help="Set the time before the screen goes blank on inactivity (0 to disable)",
        )
        timeout_parser.add_argument(
            "timeout_value", help="Timeout value", type=int, nargs="?"
        )

    @classmethod
    def add_brightness_arguments(cls, parser):
        parser.add_argument(
            "brightness_value",
            help="Set screen brightness level [1-10] on pi-topHUB, or [1-16] or pi-topHUB v2",
            type=int,
            choices=range(1, 17),
            nargs="?",
        )
        parser.add_argument(
            "-i",
            "--increment_brightness",
            help="Increment screen brightness level",
            action="store_true",
        )
        parser.add_argument(
            "-d",
            "--decrement_brightness",
            help="Decrement screen brightness level",
            action="store_true",
        )


def main():
    from .deprecated_cli_runner import run

    run(DisplayCLI)


def brightness():
    parser = ArgumentParser(
        "brightness",
        description="Control onboard pi-top display brightness (Original pi-top, pi-topCEED, pi-top [3])",
    )
    DisplayCLI.add_brightness_arguments(parser)
    args = parser.parse_args()
    args = vars(args)
    args["display_subcommand"] = "brightness"

    from .deprecated_cli_runner import run_with_args

    run_with_args(
        DisplayCLI,
        old_command="pt-brightness",
        new_command="pi-top display brightness",
        args_dict=args,
    )


if __name__ == "__main__":
    main()



from time import sleep

from pitop.common.bitwise_ops import split_into_bytes
from pitop.common.i2c_device import I2CDevice


class ADCProbe:

    __device_address = 0x2A
    __register_address = 0
    __channel_count = 6
    __adc_ratio = 20

    def __init__(self, i2c_device_name="/dev/i2c-1"):

        self.__i2c_device_name = i2c_device_name
        self.__device = None
        self.__error_array = [-1] * self.__channel_count

    def read_value(self, channel):

        if (channel < 0) or (channel > self.__channel_count):
            print("Invalid channel - use 0 to " + str(self.__channel_count - 1))
            return -1

        results = self.read_all()

        return results[channel]

    def read_all(self):

        if self.__connect() is False:
            print("Could not connect to device")
            return self.__error_array

        results_reading = self.__device.read_n_unsigned_bytes(
            self.__register_address, number_of_bytes=self.__channel_count
        )
        results = split_into_bytes(
            results_reading, self.__channel_count, little_endian=False
        )
        data_read_len = len(results)
        self.__disconnect()

        if data_read_len != self.__channel_count:
            print(
                "Bad read from device. "
                f"Expected {str(self.__channel_count)} bytes, received {str(data_read_len)} bytes."
            )
            return self.__error_array

        for i in range(self.__channel_count):
            results[i] *= self.__adc_ratio
            results[i] = int(results[i])

        return results

    def poll(self, delay=0.5):

        print("| ADC0\t| ADC1\t| ADC2\t| ADC3\t| ADC4\t| ADC5\t|")
        print("-------------------------------------------------")

        while True:

            results = self.read_all()

            print(
                "| "
                + str(results[0])
                + "\t| "
                + str(results[1])
                + "\t| "
                + str(results[2])
                + "\t| "
                + str(results[3])
                + "\t| "
                + str(results[4])
                + "\t| "
                + str(results[5])
                + "\t|"
            )
            sleep(delay)

    def __connect(self):

        try:
            self.__device = I2CDevice(self.__i2c_device_name, self.__device_address)
            self.__device.connect()

        except Exception as e:
            print("Unable to read from ADC over I2C: " + str(e))
            return False

        return True

    def __disconnect(self):

        if self.__device is not None:
            self.__device.disconnect()



from pitop.pma.adc_base import ADCBase


class SoundSensor(ADCBase):
    """Encapsulates the behaviour of a sound sensor.

    A sound sensor component is typically a simple microphone that detects the vibrations
    of the air entering the sensor and produces an analog reading based on the amplitude
    of these vibrations.

    Uses an Analog-to-Digital Converter (ADC) to turn the analog reading from the sensor
    into a digital value.

    :param str port_name: The ID for the port to which this component is connected
    """

    def __init__(self, port_name, pin_number=1, name="sound_sensor"):
        ADCBase.__init__(self, port_name=port_name, pin_number=pin_number, name=name)

    @property
    def reading(self):
        """Take a reading from the sensor.

        :return: A value representing the volume of sound detected by the sensor at the current time
        :rtype: float
        """
        reading = self.read(peak_detection=True) / 2
        return reading

    @property
    def value(self):
        """Get a simple binary value based on a reading from the device.

        :return: 1 if the sensor is detecting any sound, 0 otherwise
        :rtype: integer
        """
        value = self.reading
        return 0 if value == 0 else 1



import logging
from enum import Enum, auto
from subprocess import CalledProcessError, run

from pitop.common.command_runner import run_command

logger = logging.getLogger(__name__)


class NotificationAction:
    def __init__(self, call_to_action_text, command_str) -> None:
        self.call_to_action_text = call_to_action_text
        self.command_str = command_str


class NotificationActionManager:
    def __init__(self):
        self.actions = list()
        self.default_action = None
        self.close_action = None

    def add_action(self, call_to_action_text, command_str) -> None:
        action = NotificationAction(call_to_action_text, command_str)
        self.actions.append(action)

    def set_default_action(self, command_str) -> None:
        default_action = NotificationAction("", command_str)
        self.default_action = default_action

    def set_close_action(self, command_str) -> None:
        close_action = NotificationAction("", command_str)
        self.close_action = close_action


class NotificationUrgencyLevel(Enum):
    low = auto()
    normal = auto()
    critical = auto()


def send_notification(
    title: str,
    text: str,
    icon_name: str = "",
    timeout: int = 0,
    app_name: str = "",
    notification_id: int = -1,
    actions_manager: NotificationActionManager = None,
    urgency_level: NotificationUrgencyLevel = None,
    capture_notification_id: bool = True,
) -> str:

    # Check that `notify-send-ng` is available, as it's not a hard dependency of the package
    try:
        run(["dpkg-query", "-l", "notify-send-ng"], capture_output=True, check=True)
    except CalledProcessError:
        raise Exception("notify-send-ng not installed")

    cmd = "/usr/bin/notify-send "
    cmd += "--print-id "
    cmd += "--expire-time=" + str(timeout) + " "

    if icon_name:
        cmd += "--icon=" + icon_name + " "

    if notification_id >= 0:
        cmd += "--replace=" + str(notification_id) + " "

    if actions_manager is not None:
        for action in actions_manager.actions:
            cmd += (
                '--action="'
                + action.call_to_action_text
                + ":"
                + action.command_str
                + '" '
            )

        if actions_manager.default_action is not None:
            cmd += (
                "--default-action=" + actions_manager.default_action.command_str + " "
            )

        if actions_manager.close_action is not None:
            cmd += "--close-action=" + actions_manager.close_action.command_str + " "

    if app_name:
        cmd += "--app-name=" + app_name + " "

    if urgency_level is not None:
        cmd += "--urgency=" + urgency_level.name + " "

    cmd += ' "' + title + '" '
    cmd += '"' + text + '"'

    logger.info("notify-send command: {}".format(cmd))

    try:
        resp_stdout = run_command(cmd, 2000, capture_output=capture_notification_id)
    except Exception as e:
        logger.warning("Failed to show message: {}".format(e))
        raise
    return resp_stdout



import logging

from numpy import asarray

from .capture_action_base import CaptureActionBase

logger = logging.getLogger(__name__)


class VideoCapture(CaptureActionBase):
    """Class that stores camera frames as a video in the given
    :data:`output_file_name`.

    :param str output_file_name: The filename into which to write the video. If not provided, the file will
    be created in the directory ~/Camera/
    :param float fps: (Advanced) The framerate to use for the captured video. Defaults to 20.0 fps
    :param tuple resolution: (Advanced) The resolution to use for the captured video. Defaults to (640, 360)
    :param str codec: (Advanced) The codec to use for the captured video. Defaults to DIVX
    """

    def __init__(self, output_file_name="", fps=20.0, resolution=None):
        from imageio import get_writer

        default_video_resolution = (640, 368)

        if output_file_name == "":
            output_directory = self._create_output_directory()
            output_file_name = self._get_output_filename(output_directory, "avi")

        if resolution is None:
            resolution = default_video_resolution

        def fix_dimension(dimension):
            if dimension % 16:
                dimension = dimension + 16 - dimension % 16
                logger.warning(f"Invalid resolution. Setting dimension to {dimension}")
            return dimension

        self.__video_resolution = [fix_dimension(x) for x in resolution]
        self.__video_output_writer = get_writer(
            output_file_name, format="avi", mode="I", fps=fps
        )

    def process(self, frame):
        self.__video_output_writer.append_data(
            asarray(frame.resize(self.__video_resolution))
        )

    def stop(self):
        if self.__video_output_writer is not None:
            self.__video_output_writer.close()
            self.__video_output_writer = None



from os import path
from pathlib import Path


def retrieve_model(model_filename, base_download_link):
    def download_file(url, output_file_path):
        import sys

        import wget

        print("Downloading model file...")
        bar = wget.bar_adaptive if hasattr(sys.stdout, "fileno") else None
        try:
            wget.download(url, output_file_path, bar=bar)
        except Exception as e:
            print(e)
            raise RuntimeError(
                "Downloading model failed, please check the model_filename is correct. If issue "
                "persists, please report it here: https://github.com/pi-top/pi-top-Python-SDK/issues"
            )
        print()
        print("Download complete!")

    def extract_file(compressed_file_path, output_file_path):
        import bz2

        print("Extracting model file...")
        with open(compressed_file_path, "rb") as source, open(
            output_file_path, "wb"
        ) as dest:
            dest.write(bz2.decompress(source.read()))
        print("Extraction complete!")

    model_dir = path.join(".config", "pi-top", "sdk", "models")
    abs_model_dir = path.join(str(Path.home()), model_dir)
    model_file_path = path.join(abs_model_dir, model_filename)

    if path.exists(model_file_path):
        return model_file_path

    print(f'Required model file ("{model_filename}") not found.')

    Path(abs_model_dir).mkdir(parents=True, exist_ok=True)

    compressed_model_filename = f"{model_filename}.bz2"
    download_file_path = path.join(abs_model_dir, compressed_model_filename)

    download_file(
        url=f"{base_download_link}{compressed_model_filename}",
        output_file_path=download_file_path,
    )
    extract_file(
        compressed_file_path=download_file_path, output_file_path=model_file_path
    )

    return model_file_path


def load_emotion_model(model_filename="emotion_classification_model_svc_v1.onnx"):
    try:
        import onnxruntime as rt
    except (ImportError, ModuleNotFoundError):
        raise ModuleNotFoundError(
            "onnxruntime Python library is not installed. You can install it by running 'sudo apt install python3-onnxruntime'."
        ) from None

    model_file_path = retrieve_model(
        model_filename=model_filename,
        base_download_link="https://github.com/pi-top/Machine-Learning-Models/raw/master/",
    )

    if path.exists(model_file_path):
        return rt.InferenceSession(model_file_path)
    else:
        raise RuntimeError(
            "Retrieving model failed, please try again. If issue persists, please report it here: "
            "https://github.com/pi-top/pi-top-Python-SDK/issues"
        )


def load_face_landmark_predictor(model_filename):
    from pitop.processing.core.vision_functions import import_dlib

    dlib = import_dlib()

    model_file_path = retrieve_model(
        model_filename=model_filename,
        base_download_link="https://github.com/davisking/dlib-models/raw/master/",
    )

    if path.exists(model_file_path):
        return dlib.shape_predictor(model_file_path)
    else:
        raise RuntimeError(
            "Retrieving model failed, please try again. If issue persists, please report it here: "
            "https://github.com/pi-top/pi-top-Python-SDK/issues"
        )



import pitop.miniscreen.oled.core.contrib.luma.core.error
import pitop.miniscreen.oled.core.contrib.luma.oled.const
from pitop.miniscreen.oled.core.contrib.luma.core.device import device


class sh1106(device):
    """Serial interface to a monochrome SH1106 OLED display.

    On creation, an initialization sequence is pumped to the display to
    properly configure it. Further control commands can then be called
    to affect the brightness and other settings.
    """

    def __init__(self, serial_interface=None, width=128, height=64, rotate=0, **kwargs):
        super(sh1106, self).__init__(
            pitop.miniscreen.oled.core.contrib.luma.oled.const.sh1106, serial_interface
        )
        self.capabilities(width, height, rotate)
        self._pages = self._h // 8

        settings = {
            (128, 128): dict(multiplex=0xFF, displayoffset=0x02),
            (128, 64): dict(multiplex=0x3F, displayoffset=0x00),
            (128, 32): dict(multiplex=0x20, displayoffset=0x0F),
        }.get((width, height))

        if settings is None:
            raise pitop.miniscreen.oled.core.contrib.luma.core.error.DeviceDisplayModeError(
                f"Unsupported display mode: {width} x {height}"
            )

        self.command(
            self._const.DISPLAYOFF,
            self._const.MEMORYMODE,
            self._const.SETHIGHCOLUMN,
            0xB0,
            0xC8,
            self._const.SETLOWCOLUMN,
            0x10,
            0x40,
            self._const.SETSEGMENTREMAP,
            self._const.NORMALDISPLAY,
            self._const.SETMULTIPLEX,
            settings["multiplex"],
            self._const.DISPLAYALLON_RESUME,
            self._const.SETDISPLAYOFFSET,
            settings["displayoffset"],
            self._const.SETDISPLAYCLOCKDIV,
            0xF0,
            self._const.SETPRECHARGE,
            0x22,
            self._const.SETCOMPINS,
            0x12,
            self._const.SETVCOMDETECT,
            0x20,
            self._const.CHARGEPUMP,
            0x14,
        )

        self.contrast(0x7F)
        self.clear()
        self.show()

    def display(self, image):
        """Takes a 1-bit :py:mod:`PIL.Image` and dumps it to the SH1106 OLED
        display.

        :param image: Image to display.
        :type image: :py:mod:`PIL.Image`
        """
        assert image.mode == self.mode
        assert image.size == self.size

        image = self.preprocess(image)

        set_page_address = 0xB0
        image_data = image.getdata()
        pixels_per_page = self.width * 8
        buf = bytearray(self.width)

        for y in range(0, int(self._pages * pixels_per_page), pixels_per_page):
            self.command(set_page_address, 0x02, 0x10)
            set_page_address += 1
            offsets = [y + self.width * i for i in range(8)]

            for x in range(self.width):
                buf[x] = (
                    (image_data[x + offsets[0]] and 0x01)
                    | (image_data[x + offsets[1]] and 0x02)
                    | (image_data[x + offsets[2]] and 0x04)
                    | (image_data[x + offsets[3]] and 0x08)
                    | (image_data[x + offsets[4]] and 0x10)
                    | (image_data[x + offsets[5]] and 0x20)
                    | (image_data[x + offsets[6]] and 0x40)
                    | (image_data[x + offsets[7]] and 0x80)
                )

            self.data(list(buf))



from .componentable import Componentable
from .recreatable import Recreatable
from .stateful import Stateful
from .supports_battery import SupportsBattery
from .supports_miniscreen import SupportsMiniscreen



from argparse import ArgumentParser

from .battery import BatteryCLI
from .cli_base import PitopCliException, PitopCliInvalidArgument
from .devices import DeviceCLI
from .display import DisplayCLI
from .imu import ImuCLI
from .oled import OledCLI
from .support import SupportCLI

lookup_dict = {
    "battery": BatteryCLI,
    "devices": DeviceCLI,
    "display": DisplayCLI,
    "support": SupportCLI,
    "imu": ImuCLI,
    "oled": OledCLI,
}


def get_parser():
    """Configures the argument parser according to the CLI classes defined in.

    ´lookup_dict´, and returns the parsed arguments.

    Returns:
        ArgumentParser: parser object
    """
    parser = ArgumentParser(prog="pi-top")
    subparsers = parser.add_subparsers(
        title="Subcommands",
        description="Set of valid subcommands to use to interface with your pi-top",
        help="valid subcommands",
        dest="subcommand",
    )

    for cli_name, cli_class in lookup_dict.items():
        class_parser = subparsers.add_parser(cli_name, help=cli_class.parser_help)
        cli_class.add_parser_arguments(class_parser)
        cli_class.parser = class_parser

    return parser


def run(args):
    """Executes the command according to the provided arguments."""
    exit_code = 1
    cli = None
    try:
        cli = lookup_dict.get(args.subcommand)
        exit_code = cli(args).run()
    except PitopCliException:
        pass
    except PitopCliInvalidArgument:
        if cli:
            cli.parser.print_help()
    except Exception as e:
        print(f"Error on pitop.run: {e}")

    return exit_code


def main():
    parser = get_parser()
    args = parser.parse_args()
    if args.subcommand is None:
        parser.print_help()
        return 1

    return run(args)



from os import listdir, path, walk
from re import search
from subprocess import check_output

from ..formatter import StdoutFormat
from .systemd_service import SystemdService


class PitopSoftware:
    def print_pt_systemd_status(self):
        services = self.get_pt_systemd_services()

        for active_state, services_arr in sorted(services.items()):
            StdoutFormat.print_line(f"{SystemdService.format_service(active_state)}")
            for service in services_arr:
                service.print()

    def get_pt_systemd_services(self):
        pt_service_names = self.get_pt_systemd_service_names()
        services = {}

        for service_name in pt_service_names:
            output = check_output(
                [
                    "systemctl",
                    "show",
                    service_name,
                    "--full",
                    "--no-legend",
                    "--lines=0",
                    "--no-pager",
                ]
            )
            output = output.decode("UTF-8")
            lines = output.split("\n")
            service = SystemdService.from_lines(lines)

            active_state = service.active_state
            if active_state not in services:
                services[active_state] = []
            services[active_state].append(service)

        return services

    def get_pt_systemd_service_names(self):
        pt_service_names = []
        for systemd_service_file in sorted(listdir("/lib/systemd/system")):
            if systemd_service_file.startswith("pt-"):
                pt_service_names.append(systemd_service_file)
        return pt_service_names

    def get_pt_installed_software(self):
        try:
            from apt import Cache
        except ModuleNotFoundError:
            # probably running in virtualenv. python3-pip is not pip installable
            print("Error: couldn't find python APT library. Skipping...")
            return []

        regex = "^pt-|-pt-|pitop|pi-top"
        apt_cache = Cache()

        pt_packages = []
        for pkg in apt_cache:
            match = search(regex, pkg.name)
            if apt_cache[pkg.name].is_installed and match:
                pt_packages.append((pkg.shortname, pkg.installed.version))
        return pt_packages

    def print_apt_sources(self):
        sources_directory = "/etc/apt/"

        for root, dirs, files in walk(sources_directory):
            for file in files:
                if file.endswith(".list"):
                    path_to_source = path.join(root, file)
                    StdoutFormat.print_line(path_to_source)

                    with open(path_to_source, "r") as f:
                        lines = f.readlines()

                    for line in lines:
                        if not line.startswith("#"):
                            StdoutFormat.print_line(f"{line.strip()}", level=2)



from enum import Enum

from .generic_action import GenericAction
from .motion_detector import MotionDetector
from .store_frame import StoreFrame
from .video_capture import VideoCapture


class CaptureActions(Enum):
    CAPTURE_SINGLE_FRAME = StoreFrame
    CAPTURE_VIDEO_TO_FILE = VideoCapture
    DETECT_MOTION = MotionDetector
    HANDLE_FRAME = GenericAction



import atexit
from collections import deque
from os import getenv
from typing import Union

import numpy as np
from imutils.video import FPS

from pitop.core import ImageFunctions
from pitop.core.data_structures import DotDict
from pitop.processing.core.vision_functions import (
    center_reposition,
    get_object_target_lock_control_angle,
    import_imutils,
    import_opencv,
    tuple_for_color_by_name,
)

VALID_COLORS = ["red", "green", "blue"]
DETECTION_POINTS_BUFFER_LENGTH = 16

cv2 = None
imutils = None


def import_libs():
    global cv2, imutils
    if cv2 is None:
        cv2 = import_opencv()
    if imutils is None:
        imutils = import_imutils()


class BallLikeness:
    def __init__(self, contour):
        import_libs()

        self.contour = contour
        self.pos, self.radius = cv2.minEnclosingCircle(self.contour)
        self.area = cv2.contourArea(self.contour)

        self.circular_likeness = cv2.matchShapes(
            self.contour, self.__circular_match_contour(), 1, 0.0
        )

        # Most likely ball is a mixture of largest area and one that is most "ball-shaped"
        self.likelihood = self.area / (self.circular_likeness + 1e-5)

    def __circular_match_contour(self):
        # Initialise empty mask
        mask_to_compare = np.zeros(
            (int(2 * self.radius), int(2 * self.radius)), dtype="uint8"
        )

        # Draw circle matching contour's position and radius
        cv2.circle(
            mask_to_compare,
            (int(self.radius), int(self.radius)),
            int(self.radius),
            255,
            -1,
        )

        return max(
            imutils.grab_contours(
                cv2.findContours(
                    mask_to_compare, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
                )
            ),
            key=cv2.contourArea,
        )


class Ball:
    MIN_BALL_RADIUS = 5
    BALL_CLOSE_RADIUS = 50

    def __init__(self, color):
        self.color = color

        # 'Holding a red ball' can produce issues
        # since some skin types are redish in color
        minimum_shape_accuracies = {"red": 0.1, "green": 0.05, "blue": 0.05}
        self.minimum_shape_accuracy = minimum_shape_accuracies[color]

        self._center_points = deque(maxlen=DETECTION_POINTS_BUFFER_LENGTH)
        self._center = None
        self._radius = 0
        self._angle = None

    @property
    def angle(self) -> float:
        """
        :return: Angle between the approximate robot chassis center and the ball in the frame. Used for input to a PID
        algorithm to align a robot chassis with the ball center (drive in a direction that keeps the ball in the middle
        of the frame).
        :rtype: float
        """
        return self._angle

    @angle.setter
    def angle(self, value: float):
        self._angle = value

    @property
    def center(self) -> tuple:
        """
        :return: (x, y) coordinates of the ball where (0, 0) is the middle of the image frame used to detect it.
        :rtype: tuple
        """
        return self._center

    @center.setter
    def center(self, value: tuple):
        self._center = value

    @property
    def radius(self) -> int:
        """
        :return: Radius of the detected ball in pixels in relation to the image frame used to detect it.
        :rtype: int
        """
        return self._radius

    @radius.setter
    def radius(self, value: int):
        self._radius = value

    @property
    def center_points(self) -> deque:
        """
        :return: deque of (x, y) coordinates for historical ball detections where (0, 0) is the top left of the frame.
        :rtype: deque
        """
        return self._center_points

    @property
    def found(self) -> bool:
        """
        :return: Boolean to determine if a valid ball was found in the frame.
        :rtype: bool
        """
        return self.center is not None

    def will_accept(self, ball_likeness: BallLikeness):
        if ball_likeness.radius < self.MIN_BALL_RADIUS:
            return False

        if ball_likeness.circular_likeness < self.minimum_shape_accuracy:
            return True

        if ball_likeness.radius > self.BALL_CLOSE_RADIUS:
            return True

        return False


class BallDetector:
    def __init__(self, image_processing_width: int = 320, format: str = "OpenCV"):
        """
        :param int image_processing_width: image width to scale to for image processing
        :param str format: output image format
        """
        import_libs()

        self._image_processing_width = image_processing_width
        self.format = format
        self.balls = {c: Ball(c) for c in VALID_COLORS}
        self._frame_scaler = None

        # Enable FPS if environment variable is set
        self._print_fps = getenv("PT_ENABLE_FPS", "0") == "1"
        if self._print_fps:
            self._fps = FPS().start()
            atexit.register(self.__print_fps)

    def __call__(self, frame, color: Union[str, list] = "red"):
        def parse_colors(color_arg):
            colors = []
            if type(color_arg) == str:
                if color_arg not in VALID_COLORS:
                    raise ValueError(
                        f"Valid color values are {', '.join(VALID_COLORS[:-1])} or {VALID_COLORS[-1]}"
                    )
                colors = [color_arg]
            elif type(color_arg) in (list, tuple):
                if not set(color_arg).issubset(VALID_COLORS):
                    raise ValueError(
                        f"Valid color values are {', '.join(VALID_COLORS[:-1])} or {VALID_COLORS[-1]}"
                    )
                colors = color_arg
            if len(colors) > 3:
                raise ValueError("Cannot pass more than three colors.")

            return colors

        import_libs()
        frame = ImageFunctions.convert(frame, format="OpenCV")

        if self._frame_scaler is None:
            _, width = frame.shape[0:2]
            self._frame_scaler = width / self._image_processing_width

        for c in parse_colors(color):
            self.balls[c] = self.__find_most_likely_ball(
                ball=self.balls.get(c), frame=frame, color=c
            )

        robot_view = frame.copy()
        ball_data = DotDict({})
        for ball_color, ball_object in self.balls.items():
            ball_data[ball_color] = ball_object
            self.__draw_ball_contrail(robot_view, ball_object)
            if ball_object.found:
                self.__draw_ball_position(robot_view, ball_object)

        ball_data["robot_view"] = ImageFunctions.convert(robot_view, self.format)

        if self._print_fps:
            self._fps.update()

        return ball_data

    def __print_fps(self):
        self._fps.stop()
        print(f"[INFO] Elapsed time: {self._fps.elapsed():.2f}")
        print(f"[INFO] Approx. FPS: {self._fps.fps():.2f}")

    def __draw_ball_position(self, frame, ball):
        cv2.circle(frame, ball.center_points[0], ball.radius, (0, 255, 255), 2)
        cv2.circle(
            frame,
            ball.center_points[0],
            5,
            tuple_for_color_by_name(ball.color, bgr=True),
            -1,
        )

    def __draw_ball_contrail(self, frame, ball):
        for i in range(1, len(ball.center_points)):
            if ball.center_points[i - 1] is None or ball.center_points[i] is None:
                continue
            thickness = int(np.sqrt(DETECTION_POINTS_BUFFER_LENGTH / float(i + 1)))

            cv2.line(
                frame,
                ball.center_points[i - 1],
                ball.center_points[i],
                tuple_for_color_by_name(ball.color, bgr=True),
                thickness,
            )

    def color_filter(self, frame, color: str = "red"):
        frame = ImageFunctions.convert(frame, format="OpenCV")
        mask = self.__get_color_mask(frame, color)
        filtered_image = cv2.bitwise_and(frame, frame, mask=mask)
        if self.format.lower() == "pil":
            filtered_image = ImageFunctions.convert(mask, format="PIL")

        return filtered_image

    def __get_color_mask(self, frame, color: str):
        blurred = cv2.blur(frame, (11, 11))
        hsv = cv2.cvtColor(blurred, cv2.COLOR_BGR2HSV)

        masks = []
        color_ranges = {
            "red": [
                {"lower": (150, 100, 100), "upper": (179, 255, 255)},
                {"lower": (0, 100, 100), "upper": (5, 255, 255)},
            ],
            "green": [{"lower": (60, 100, 100), "upper": (90, 255, 255)}],
            "blue": [{"lower": (100, 100, 100), "upper": (130, 255, 255)}],
        }
        for color_range in color_ranges[color]:
            hsv_lower = color_range["lower"]
            hsv_upper = color_range["upper"]
            mask = cv2.inRange(hsv, hsv_lower, hsv_upper)
            masks.append(mask)
        mask = sum(masks)

        mask = cv2.erode(mask, None, iterations=1)
        mask = cv2.dilate(mask, None, iterations=1)

        return mask

    def __find_contours(self, frame, color):
        mask = self.__get_color_mask(frame, color=color)

        return (
            imutils.grab_contours(  # fixes problems with OpenCV changing their protocol
                cv2.findContours(
                    mask.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
                )
            )
        )

    def __find_most_likely_ball(self, ball, frame, color):
        resized_frame = imutils.resize(frame, width=self._image_processing_width)
        contours = self.__find_contours(resized_frame, color)
        if len(contours) == 0:
            ball.center_points.appendleft(None)
            ball.center = None
            ball.radius = 0
            ball.angle = None

            return ball

        highest_likelihood = 0

        ball_center = None
        ball_radius = 0
        for contour in contours:
            ball_likeness = BallLikeness(contour)

            if ball_likeness.likelihood < highest_likelihood:
                continue

            # Don't accept future balls which are less likely
            highest_likelihood = ball_likeness.likelihood
            if ball.will_accept(ball_likeness):
                # Scale to original frame size

                ball_center = tuple(
                    int(coordinate * self._frame_scaler)
                    for coordinate in ball_likeness.pos
                )
                ball_radius = int(ball_likeness.radius * self._frame_scaler)

        # Update ball state
        ball.center_points.appendleft(ball_center)
        ball.center = (
            center_reposition(ball_center, frame) if ball_center is not None else None
        )
        ball.angle = (
            get_object_target_lock_control_angle(ball.center, frame)
            if ball_center is not None
            else None
        )
        ball.radius = ball_radius

        return ball



import time
import warnings


def _clamp(value, limits):
    lower, upper = limits
    if value is None:
        return None
    elif (upper is not None) and (value > upper):
        return upper
    elif (lower is not None) and (value < lower):
        return lower
    return value


try:
    # Get monotonic time to ensure that time deltas are always positive
    _current_time = time.monotonic
except AttributeError:
    # time.monotonic() not available (using python < 3.3), fallback to time.time()
    _current_time = time.time
    warnings.warn(
        "time.monotonic() not available in python < 3.3, using time.time() as fallback"
    )


class PID(object):
    """A simple PID controller."""

    def __init__(
        self,
        Kp=1.0,
        Ki=0.0,
        Kd=0.0,
        setpoint=0,
        sample_time=0.01,
        output_limits=(None, None),
        auto_mode=True,
        proportional_on_measurement=False,
        error_map=None,
    ):
        """Initialize a new PID controller.

        :param Kp: The value for the proportional gain Kp
        :param Ki: The value for the integral gain Ki
        :param Kd: The value for the derivative gain Kd
        :param setpoint: The initial setpoint that the PID will try to achieve
        :param sample_time: The time in seconds which the controller should wait before generating
            a new output value. The PID works best when it is constantly called (eg. during a
            loop), but with a sample time set so that the time difference between each update is
            (close to) constant. If set to None, the PID will compute a new output value every time
            it is called.
        :param output_limits: The initial output limits to use, given as an iterable with 2
            elements, for example: (lower, upper). The output will never go below the lower limit
            or above the upper limit. Either of the limits can also be set to None to have no limit
            in that direction. Setting output limits also avoids integral windup, since the
            integral term will never be allowed to grow outside of the limits.
        :param auto_mode: Whether the controller should be enabled (auto mode) or not (manual mode)
        :param proportional_on_measurement: Whether the proportional term should be calculated on
            the input directly rather than on the error (which is the traditional way). Using
            proportional-on-measurement avoids overshoot for some types of systems.
        :param error_map: Function to transform the error value in another constrained value.
        """
        self.Kp, self.Ki, self.Kd = Kp, Ki, Kd
        self.setpoint = setpoint
        self.sample_time = sample_time

        self._min_output, self._max_output = None, None
        self._auto_mode = auto_mode
        self.proportional_on_measurement = proportional_on_measurement
        self.error_map = error_map

        self._proportional = 0
        self._integral = 0
        self._derivative = 0

        self._last_time = None
        self._last_output = None
        self._last_input = None

        self.output_limits = output_limits
        self.reset()

    def __call__(self, input_, dt=None):
        """Update the PID controller.

        Call the PID controller with *input_* and calculate and return a control output if
        sample_time seconds has passed since the last update. If no new output is calculated,
        return the previous output instead (or None if no value has been calculated yet).

        :param dt: If set, uses this value for timestep instead of real time. This can be used in
            simulations when simulation time is different from real time.
        """
        if not self.auto_mode:
            return self._last_output

        now = _current_time()
        if dt is None:
            dt = now - self._last_time if (now - self._last_time) else 1e-16
        elif dt <= 0:
            raise ValueError("dt has negative value {}, must be positive".format(dt))

        if (
            self.sample_time is not None
            and dt < self.sample_time
            and self._last_output is not None
        ):
            # Only update every sample_time seconds
            return self._last_output

        # Compute error terms
        error = self.setpoint - input_
        d_input = input_ - (
            self._last_input if (self._last_input is not None) else input_
        )

        # Check if must map the error
        if self.error_map is not None:
            error = self.error_map(error)

        # Compute the proportional term
        if not self.proportional_on_measurement:
            # Regular proportional-on-error, simply set the proportional term
            self._proportional = self.Kp * error
        else:
            # Add the proportional error on measurement to error_sum
            self._proportional -= self.Kp * d_input

        # Compute integral and derivative terms
        self._integral += self.Ki * error * dt
        self._integral = _clamp(
            self._integral, self.output_limits
        )  # Avoid integral windup

        self._derivative = -self.Kd * d_input / dt

        # Compute final output
        output = self._proportional + self._integral + self._derivative
        output = _clamp(output, self.output_limits)

        # Keep track of state
        self._last_output = output
        self._last_input = input_
        self._last_time = now

        return output

    def __repr__(self):
        return (
            "{self.__class__.__name__}("
            "Kp={self.Kp!r}, Ki={self.Ki!r}, Kd={self.Kd!r}, "
            "setpoint={self.setpoint!r}, sample_time={self.sample_time!r}, "
            "output_limits={self.output_limits!r}, auto_mode={self.auto_mode!r}, "
            "proportional_on_measurement={self.proportional_on_measurement!r},"
            "error_map={self.error_map!r}"
            ")"
        ).format(self=self)

    @property
    def components(self):
        """The P-, I- and D-terms from the last computation as separate
        components as a tuple.

        Useful for visualizing what the controller is doing or when
        tuning hard-to-tune systems.
        """
        return self._proportional, self._integral, self._derivative

    @property
    def tunings(self):
        """The tunings used by the controller as a tuple: (Kp, Ki, Kd)."""
        return self.Kp, self.Ki, self.Kd

    @tunings.setter
    def tunings(self, tunings):
        """Set the PID tunings."""
        self.Kp, self.Ki, self.Kd = tunings

    @property
    def auto_mode(self):
        """Whether the controller is currently enabled (in auto mode) or
        not."""
        return self._auto_mode

    @auto_mode.setter
    def auto_mode(self, enabled):
        """Enable or disable the PID controller."""
        self.set_auto_mode(enabled)

    def set_auto_mode(self, enabled, last_output=None):
        """Enable or disable the PID controller, optionally setting the last
        output value.

        This is useful if some system has been manually controlled and if the PID should take over.
        In that case, disable the PID by setting auto mode to False and later when the PID should
        be turned back on, pass the last output variable (the control variable) and it will be set
        as the starting I-term when the PID is set to auto mode.

        :param enabled: Whether auto mode should be enabled, True or False
        :param last_output: The last output, or the control variable, that the PID should start
            from when going from manual mode to auto mode. Has no effect if the PID is already in
            auto mode.
        """
        if enabled and not self._auto_mode:
            # Switching from manual mode to auto, reset
            self.reset()

            self._integral = last_output if (last_output is not None) else 0
            self._integral = _clamp(self._integral, self.output_limits)

        self._auto_mode = enabled

    @property
    def output_limits(self):
        """The current output limits as a 2-tuple: (lower, upper).

        See also the *output_limits* parameter in :meth:`PID.__init__`.
        """
        return self._min_output, self._max_output

    @output_limits.setter
    def output_limits(self, limits):
        """Set the output limits."""
        if limits is None:
            self._min_output, self._max_output = None, None
            return

        min_output, max_output = limits

        if (None not in limits) and (max_output < min_output):
            raise ValueError("lower limit must be less than upper limit")

        self._min_output = min_output
        self._max_output = max_output

        self._integral = _clamp(self._integral, self.output_limits)
        self._last_output = _clamp(self._last_output, self.output_limits)

    def reset(self):
        """Reset the PID controller internals.

        This sets each term to 0 as well as clearing the integral, the
        last output and the last input (derivative calculation).
        """
        self._proportional = 0
        self._integral = 0
        self._derivative = 0

        self._integral = _clamp(self._integral, self.output_limits)

        self._last_time = _current_time()
        self._last_output = None
        self._last_input = None



import subprocess
from fractions import Fraction
from os import path, uname

import netifaces
from isc_dhcp_leases import IscDhcpLeases

from pitop.common.command_runner import run_command

_, _, _, _, machine = uname()


def is_pi():
    return machine in ("armv7l", "aarch64")


def get_uname_release():
    return uname().release


def get_uname_version():
    return uname().version


def get_debian_version():
    debian_version_file = "/etc/debian_version"
    if not path.exists(debian_version_file):
        return None
    with open(debian_version_file, "r") as reader:
        content = reader.read()
    return content.strip()


def get_maj_debian_version():
    version = None
    with open("/etc/os-release", "r") as f:
        for line in f:
            if "VERSION_ID=" in line:
                quote_wrapped_version = line.split("=")[1]
                version_str = quote_wrapped_version.replace('"', "").replace("\n", "")
                try:
                    version = int(version_str)
                except ValueError:
                    version = None
                break

    try:
        return int(version)
    except ValueError:
        return None


def get_network_strength(iface):
    strength = -1
    try:
        response_str = str(subprocess.check_output(["iwconfig", iface]).decode("utf-8"))
        response_lines = response_str.splitlines()
        for line in response_lines:
            if "Link Quality" in line:
                strength_str = line.lstrip(" ").lstrip("Link Quality=").split(" ")[0]
                strength = int(Fraction(strength_str) * 100)
                break
    except (FileNotFoundError, subprocess.CalledProcessError):
        pass

    return str(strength) + "%"


def get_wifi_network_ssid():
    try:
        network_id = str(
            subprocess.check_output(["iwgetid", "-r"]).decode("utf-8")
        ).strip()
    except (FileNotFoundError, subprocess.CalledProcessError):
        network_id = "Error"

    return network_id


def get_internal_ip(iface="wlan0"):
    if iface not in netifaces.interfaces():
        return iface + " Not Found"

    try:
        addrs = netifaces.ifaddresses(iface)
    except Exception:
        return "Addresses Not Found"

    try:
        inet_addrs = addrs[netifaces.AF_INET][0]
    except Exception:
        return "Internet Addresses Not Found"

    try:
        internal_ip = inet_addrs["addr"]
    except Exception:
        return "IP Not Found"

    return internal_ip


def start_systemd_service(service_name: str):
    try:
        run_command(
            f"systemctl start {service_name}", timeout=20, check=True, log_errors=False
        )
    except Exception:
        pass


def stop_systemd_service(service_name: str):
    try:
        run_command(
            f"systemctl stop {service_name}", timeout=20, check=False, log_errors=False
        )
    except Exception:
        pass


def get_systemd_active_state(service_name: str):
    try:
        state = run_command(
            f"systemctl is-active {service_name}", timeout=10, log_errors=False
        )
        state = str(state.strip())
    except Exception:
        state = "Unknown Error"
    finally:
        return state


def get_systemd_enabled_state(service_to_check: str):
    try:
        state = str(
            subprocess.check_output(
                ["systemctl", "is-enabled", service_to_check]
            ).decode("utf-8")
        )
    except subprocess.CalledProcessError as response:
        state = str(response.output.decode("utf-8"))
    except Exception:
        state = "Unknown Error"
    finally:
        return state.strip().capitalize()


def get_ssh_enabled_state():
    ssh_enabled_state = get_systemd_enabled_state("ssh")
    return ssh_enabled_state


def get_vnc_enabled_state():
    vnc_enabled_state = get_systemd_enabled_state("vncserver-x11-serviced.service")
    return vnc_enabled_state


def get_pt_further_link_enabled_state():
    vnc_enabled_state = get_systemd_enabled_state("further-link.service")
    return vnc_enabled_state


def get_ap_mode_status():
    key_lookup = {
        "State": "state",
        "Access Point Network SSID": "ssid",
        "Access Point Wi-Fi Password": "passphrase",
        "Access Point IP Address": "ip_address",
    }

    data = {}
    try:
        ap_mode_status = run_command("/usr/bin/wifi-ap-sta status", timeout=10)
        for str in ap_mode_status.strip().split("\n"):
            k, v = str.split(":")
            key = key_lookup.get(k.strip())
            if key:
                data[key] = v.strip()
    except Exception:
        pass
    return data


def interface_is_up(interface_name):
    operstate_file = "/sys/class/net/" + interface_name + "/operstate"
    if not path.exists(operstate_file):
        return False

    contents = ""
    with open(operstate_file, "r") as file:
        contents = file.read()

    return "up" in contents


def get_address_for_connected_device():
    def command_succeeds(cmd, timeout):
        try:
            run_command(cmd, timeout=timeout, check=True, log_errors=False)
            return True
        except Exception:
            return False

    current_leases = IscDhcpLeases("/var/lib/dhcp/dhcpd.leases").get_current().values()
    current_leases = list(current_leases)
    current_leases.reverse()

    for lease in current_leases:
        # Windows machines won't respond to ping requests by default. Using arping
        # helps us on that case, but since it takes ~1.5s, it's used as a fallback
        if command_succeeds(f"ping -c1 {lease.ip}", 0.1) or command_succeeds(
            f"arping -c1 {lease.ip}", 2
        ):
            return lease.ip

    return ""


def get_address_for_ptusb_connected_device():
    if interface_is_up("ptusb0"):
        return get_address_for_connected_device()
    return ""


def is_connected_to_internet() -> bool:
    try:
        run_command("ping -c1 8.8.8.8", timeout=2, check=True, log_errors=False)
        return True
    except Exception:
        return False



class RegisterTypes:
    ACC = 0
    GYRO = 1
    MAG = 2
    ORIENTATION = 3


class RawRegisterTypes:
    X = 0
    Y = 1
    Z = 2


class OrientationRegisterTypes:
    ROLL = 0
    PITCH = 1
    YAW = 2


class MagCalRegisterTypes:
    HARD = 0
    SOFT = 1


class MagCalHardTypes:
    X = 0
    Y = 1
    Z = 2


class MagCalSoftTypes:
    XX = 0
    YY = 1
    ZZ = 2
    XY = 3
    XZ = 4
    YZ = 5


ImuEnableRegisters = {
    RegisterTypes.ACC: 0x90,
    RegisterTypes.GYRO: 0x91,
    RegisterTypes.MAG: 0x92,
    RegisterTypes.ORIENTATION: 0x93,
}

ImuConfigRegisters = {RegisterTypes.ACC: 0xA0, RegisterTypes.GYRO: 0xA2}

MagCalibrationRegisters = {
    MagCalRegisterTypes.HARD: {
        MagCalHardTypes.X: 0xB0,
        MagCalHardTypes.Y: 0xB1,
        MagCalHardTypes.Z: 0xB2,
    },
    MagCalRegisterTypes.SOFT: {
        MagCalSoftTypes.XX: 0xB3,
        MagCalSoftTypes.YY: 0xB4,
        MagCalSoftTypes.ZZ: 0xB5,
        MagCalSoftTypes.XY: 0xB6,
        MagCalSoftTypes.XZ: 0xB7,
        MagCalSoftTypes.YZ: 0xB8,
    },
}

ScaleMappings = {
    RegisterTypes.ACC: {2: 0x00, 4: 0x01, 8: 0x02, 16: 0x03},
    RegisterTypes.GYRO: {250: 0x00, 500: 0x01, 1000: 0x02, 2000: 0x03},
}

ImuDataRegisters = {
    RegisterTypes.ACC: {
        RawRegisterTypes.X: 0x80,
        RawRegisterTypes.Y: 0x81,
        RawRegisterTypes.Z: 0x82,
    },
    RegisterTypes.GYRO: {
        RawRegisterTypes.X: 0x83,
        RawRegisterTypes.Y: 0x84,
        RawRegisterTypes.Z: 0x85,
    },
    RegisterTypes.MAG: {
        RawRegisterTypes.X: 0x86,
        RawRegisterTypes.Y: 0x87,
        RawRegisterTypes.Z: 0x88,
    },
    RegisterTypes.ORIENTATION: {
        OrientationRegisterTypes.ROLL: 0x89,
        OrientationRegisterTypes.PITCH: 0x8A,
        OrientationRegisterTypes.YAW: 0x8B,
    },
}


class ImuRegisters:
    ENABLE = ImuEnableRegisters
    DATA = ImuDataRegisters
    CONFIG = ImuConfigRegisters
    MAGCAL = MagCalibrationRegisters



def import_opencv():
    try:
        import cv2

        return cv2
    except (ImportError, ModuleNotFoundError):
        raise ModuleNotFoundError(
            "OpenCV Python library is not installed. You can install it by running 'sudo apt install python3-opencv libatlas-base-dev'."
        ) from None



from io import BytesIO

import gevent
from flask import Blueprint, Response, abort


class VideoResponse(Response):
    def __init__(self, get_frame=None, *args, **kwargs):
        if get_frame is None:
            abort(500, "Unable to get frames")

        def _get_frame():
            try:
                frame = get_frame()
                buffered = BytesIO()
                frame.save(buffered, format="JPEG")
                return buffered.getvalue()
            except Exception as e:
                print(e)

        def generate_frames():
            pool = gevent.get_hub().threadpool
            while True:
                # get_frame in thread so it won't block handler greenlets
                frame_bytes = pool.spawn(_get_frame).get()
                yield (
                    b"--frame\r\n"
                    b"Content-Type: image/jpeg\r\n\r\n" + frame_bytes + b"\r\n"
                )

        Response.__init__(
            self,
            generate_frames(),
            mimetype="multipart/x-mixed-replace; boundary=frame",
            **kwargs,
        )


class VideoBlueprint(Blueprint):
    def __init__(self, name="video", get_frame=None, **kwargs):
        Blueprint.__init__(
            self,
            name,
            __name__,
            static_folder="video",
            template_folder="templates",
            **kwargs,
        )

        @self.route(f"/{name}.mjpg")
        def video():
            return VideoResponse(get_frame=get_frame)



from typing import Tuple

from pitop.common.bitwise_ops import join_bytes, split_into_bytes

from .common import type_check
from .common.encoder_motor_registers import (
    MotorControlModes,
    MotorControlRegisters,
    MotorRegisterTypes,
)
from .plate_interface import PlateInterface


class EncoderMotorController:
    """Class used to read/write motor encoder registers from the MCU."""

    _MAX_DC_MOTOR_RPM = 6000

    def __init__(self, port: str, braking_type: int = 0) -> None:
        if port not in MotorControlRegisters.__members__:
            raise Exception("Invalid port. Motors must be connected to ports M0-M3")

        self._mcu_device = PlateInterface().get_device_mcu()
        self.registers = MotorControlRegisters[port].value
        self.set_braking_type(braking_type)

    def control_mode(self) -> MotorControlModes:
        reported_control_mode = self._mcu_device.read_unsigned_byte(
            self.registers[MotorRegisterTypes.CONTROL_MODE]
        )
        return MotorControlModes(reported_control_mode)

    @type_check
    def set_control_mode(self, mode: MotorControlModes) -> None:
        self._mcu_device.write_byte(
            self.registers[MotorRegisterTypes.CONTROL_MODE], mode.value
        )

    def braking_type(self) -> int:
        return self._mcu_device.read_unsigned_byte(
            self.registers[MotorRegisterTypes.BRAKE_TYPE]
        )

    @type_check
    def set_braking_type(self, brake_type: int) -> None:
        if brake_type not in (0, 1):
            raise ValueError("Brake type must be 0 or 1")

        self._mcu_device.write_byte(
            self.registers[MotorRegisterTypes.BRAKE_TYPE], brake_type
        )

    def tachometer(self) -> int:
        tachometer_data = 99999
        while abs(tachometer_data) > self._MAX_DC_MOTOR_RPM:
            tachometer_data = self._mcu_device.read_signed_word(
                self.registers[MotorRegisterTypes.TACHOMETER], little_endian=True
            )

        return tachometer_data

    def odometer(self) -> int:
        rotation_counter = self._mcu_device.read_n_signed_bytes(
            self.registers[MotorRegisterTypes.ODOMETER], 4, little_endian=True
        )

        return rotation_counter

    @type_check
    def set_odometer(self, rotations: int) -> None:
        list_to_send = split_into_bytes(rotations, 4, signed=True, little_endian=True)
        self._mcu_device.write_n_bytes(
            self.registers[MotorRegisterTypes.ODOMETER], list_to_send
        )

    def stop(self):
        current_control_mode = self.control_mode()

        if current_control_mode == MotorControlModes.MODE_0:
            self.set_power(0)
        elif current_control_mode == MotorControlModes.MODE_1:
            self.set_rpm_control(0)
        elif current_control_mode == MotorControlModes.MODE_2:
            self.set_rpm_with_rotations(0, 0)

    # -------------------------------------------------------------------------------------------------------------------
    # CONTROL MODE 0

    def power(self) -> int:
        if self.control_mode() != MotorControlModes.MODE_0:
            return None

        return self._mcu_device.read_signed_word(
            self.registers[MotorRegisterTypes.MODE_0_POWER], little_endian=True
        )

    @type_check
    def set_power(self, power: int) -> None:
        if power < -1000 or power > 1000:
            raise ValueError(
                "Power value sent to Expansion Plate needs to be from -1000 to +1000"
            )

        self.set_control_mode(MotorControlModes.MODE_0)
        self._mcu_device.write_word(
            self.registers[MotorRegisterTypes.MODE_0_POWER],
            power,
            little_endian=True,
            signed=True,
        )

    # -------------------------------------------------------------------------------------------------------------------
    # CONTROL MODE 1

    def rpm_control(self) -> int:
        if self.control_mode() != MotorControlModes.MODE_1:
            return None

        return self._mcu_device.read_signed_word(
            self.registers[MotorRegisterTypes.MODE_1_RPM], little_endian=True
        )

    @type_check
    def set_rpm_control(self, rpm: int) -> None:
        self.set_control_mode(MotorControlModes.MODE_1)
        self._mcu_device.write_word(
            self.registers[MotorRegisterTypes.MODE_1_RPM],
            rpm,
            signed=True,
            little_endian=True,
        )

    # -------------------------------------------------------------------------------------------------------------------
    # CONTROL MODE 2

    def rpm_with_rotations(self) -> Tuple[int, int]:
        if self.control_mode() != MotorControlModes.MODE_2:
            return None

        value = self._mcu_device.read_n_unsigned_bytes(
            self.registers[MotorRegisterTypes.MODE_2_RPM_WITH_ROTATIONS],
            4,
            little_endian=True,
        )
        value_in_bytes = split_into_bytes(value, 4, little_endian=True)

        speed = join_bytes(value_in_bytes[0:2])
        rotations = join_bytes(value_in_bytes[2:4])

        return speed, rotations

    @type_check
    def set_rpm_with_rotations(self, rpm: int, rotations_to_complete: int) -> None:
        self.set_control_mode(MotorControlModes.MODE_2)
        list_to_send = split_into_bytes(
            rotations_to_complete, 2, signed=True, little_endian=True
        ) + split_into_bytes(rpm, 2, signed=True, little_endian=True)
        self._mcu_device.write_n_bytes(
            self.registers[MotorRegisterTypes.MODE_2_RPM_WITH_ROTATIONS], list_to_send
        )



# coding=utf-8
# pynput
# Copyright (C) 2015-2018 Moses Palmér
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""The keyboard implementation for *Xorg*."""

# pylint: disable=C0111
# The documentation is extracted from the base classes

# pylint: disable=R0903
# We implement stubs

import enum
import threading

import Xlib.display
import Xlib.ext
import Xlib.ext.xtest
import Xlib.keysymdef.xkb
import Xlib.protocol
import Xlib.X
import Xlib.XK
from pynput._util import NotifierMixin
from pynput._util.xorg import (
    ListenerMixin,
    alt_gr_mask,
    alt_mask,
    display_manager,
    index_to_shift,
    keyboard_mapping,
    numlock_mask,
    shift_to_index,
    symbol_to_keysym,
)
from pynput._util.xorg_keysyms import CHARS, DEAD_KEYS, KEYPAD_KEYS, KEYSYMS, SYMBOLS

from . import _base


class KeyCode(_base.KeyCode):
    @classmethod
    def _from_symbol(cls, symbol, **kwargs):
        """Creates a key from a symbol.

        :param str symbol: The symbol name.

        :return: a key code
        """
        # First try simple translation
        keysym = Xlib.XK.string_to_keysym(symbol)
        if keysym:
            return cls.from_vk(keysym, **kwargs)

        # If that fails, try checking a module attribute of Xlib.keysymdef.xkb
        if not keysym:
            # pylint: disable=W0702; we want to ignore errors
            try:
                return cls.from_vk(
                    getattr(Xlib.keysymdef.xkb, "XK_" + symbol, 0), **kwargs
                )
            except Exception:
                return cls.from_vk(SYMBOLS.get(symbol, (0,))[0], **kwargs)
            # pylint: enable=W0702


class Key(enum.Enum):
    # Default keys
    alt = KeyCode._from_symbol("Alt_L")
    alt_l = KeyCode._from_symbol("Alt_L")
    alt_r = KeyCode._from_symbol("Alt_R")
    alt_gr = KeyCode._from_symbol("Mode_switch")
    backspace = KeyCode._from_symbol("BackSpace")
    caps_lock = KeyCode._from_symbol("Caps_Lock")
    cmd = KeyCode._from_symbol("Super_L")
    cmd_l = KeyCode._from_symbol("Super_L")
    cmd_r = KeyCode._from_symbol("Super_R")
    ctrl = KeyCode._from_symbol("Control_L")
    ctrl_l = KeyCode._from_symbol("Control_L")
    ctrl_r = KeyCode._from_symbol("Control_R")
    delete = KeyCode._from_symbol("Delete")
    down = KeyCode._from_symbol("Down")
    end = KeyCode._from_symbol("End")
    enter = KeyCode._from_symbol("Return")
    esc = KeyCode._from_symbol("Escape")
    f1 = KeyCode._from_symbol("F1")
    f2 = KeyCode._from_symbol("F2")
    f3 = KeyCode._from_symbol("F3")
    f4 = KeyCode._from_symbol("F4")
    f5 = KeyCode._from_symbol("F5")
    f6 = KeyCode._from_symbol("F6")
    f7 = KeyCode._from_symbol("F7")
    f8 = KeyCode._from_symbol("F8")
    f9 = KeyCode._from_symbol("F9")
    f10 = KeyCode._from_symbol("F10")
    f11 = KeyCode._from_symbol("F11")
    f12 = KeyCode._from_symbol("F12")
    f13 = KeyCode._from_symbol("F13")
    f14 = KeyCode._from_symbol("F14")
    f15 = KeyCode._from_symbol("F15")
    f16 = KeyCode._from_symbol("F16")
    f17 = KeyCode._from_symbol("F17")
    f18 = KeyCode._from_symbol("F18")
    f19 = KeyCode._from_symbol("F19")
    f20 = KeyCode._from_symbol("F20")
    home = KeyCode._from_symbol("Home")
    left = KeyCode._from_symbol("Left")
    page_down = KeyCode._from_symbol("Page_Down")
    page_up = KeyCode._from_symbol("Page_Up")
    right = KeyCode._from_symbol("Right")
    shift = KeyCode._from_symbol("Shift_L")
    shift_l = KeyCode._from_symbol("Shift_L")
    shift_r = KeyCode._from_symbol("Shift_R")
    space = KeyCode._from_symbol("space", char=" ")
    tab = KeyCode._from_symbol("Tab")
    up = KeyCode._from_symbol("Up")

    insert = KeyCode._from_symbol("Insert")
    menu = KeyCode._from_symbol("Menu")
    num_lock = KeyCode._from_symbol("Num_Lock")
    pause = KeyCode._from_symbol("Pause")
    print_screen = KeyCode._from_symbol("Print")
    scroll_lock = KeyCode._from_symbol("Scroll_Lock")


class Controller(NotifierMixin, _base.Controller):
    _KeyCode = KeyCode
    _Key = Key

    #: The shift mask for :attr:`Key.ctrl`
    CTRL_MASK = Xlib.X.ControlMask

    #: The shift mask for :attr:`Key.shift`
    SHIFT_MASK = Xlib.X.ShiftMask

    def __init__(self, *args, **kwargs):
        super(Controller, self).__init__(*args, **kwargs)
        self._display = Xlib.display.Display()
        self._keyboard_mapping = None
        self._borrows = {}
        self._borrow_lock = threading.RLock()

        # pylint: disable=C0103; this is treated as a class scope constant, but
        # we cannot set it in the class scope, as it requires a Display instance
        self.ALT_MASK = alt_mask(self._display)
        self.ALT_GR_MASK = alt_gr_mask(self._display)
        # pylint: enable=C0103

    def __del__(self):
        if self._display:
            self._display.close()

    @property
    def keyboard_mapping(self):
        """A mapping from *keysyms* to *key codes*.

        Each value is the tuple ``(key_code, shift_state)``. By sending an
        event with the specified *key code* and shift state, the specified
        *keysym* will be touched.
        """
        if not self._keyboard_mapping:
            self._update_keyboard_mapping()
        return self._keyboard_mapping

    def _handle(self, key, is_press):
        """Resolves a key identifier and sends a keyboard event.

        :param event: The *X* keyboard event.

        :param int keysym: The keysym to handle.
        """
        event = (
            Xlib.display.event.KeyPress if is_press else Xlib.display.event.KeyRelease
        )
        keysym = self._keysym(key)

        # Make sure to verify that the key was resolved
        if keysym is None:
            raise self.InvalidKeyException(key)

        # If the key has a virtual key code, use that immediately with
        # fake_input; fake input,being an X server extension, has access to more
        # internal state that we
        if key.vk is not None:
            with display_manager(self._display) as dm:
                Xlib.ext.xtest.fake_input(
                    dm,
                    Xlib.X.KeyPress if is_press else Xlib.X.KeyRelease,
                    dm.keysym_to_keycode(key.vk),
                )

        # Otherwise use XSendEvent; we need to use this in the general case to
        # work around problems with keyboard layouts
        else:
            try:
                keycode, shift_state = self.keyboard_mapping[keysym]
                self._send_key(event, keycode, shift_state)

            except KeyError:
                with self._borrow_lock:
                    keycode, index, count = self._borrows[keysym]
                    self._send_key(event, keycode, index_to_shift(self._display, index))
                    count += 1 if is_press else -1
                    self._borrows[keysym] = (keycode, index, count)

        # Notify any running listeners
        self._emit("_on_fake_event", key, is_press)

    def _keysym(self, key):
        """Converts a key to a *keysym*.

        :param KeyCode key: The key code to convert.
        """
        return (
            self._resolve_dead(key)
            if key.is_dead
            else None
            or self._resolve_special(key)
            or self._resolve_normal(key)
            or self._resolve_borrowed(key)
            or self._resolve_borrowing(key)
        )

    def _send_key(self, event, keycode, shift_state):
        """Sends a single keyboard event.

        :param event: The *X* keyboard event.

        :param int keycode: The calculated keycode.

        :param int shift_state: The shift state. The actual value used is
            :attr:`shift_state` or'd with this value.
        """
        with display_manager(self._display) as dm, self.modifiers as modifiers:
            # Under certain cimcumstances, such as when running under Xephyr,
            # the value returned by dm.get_input_focus is an int
            window = dm.get_input_focus().focus
            send_event = getattr(
                window, "send_event", lambda event: dm.send_event(window, event)
            )
            send_event(
                event(
                    detail=keycode,
                    state=shift_state | self._shift_mask(modifiers),
                    time=0,
                    root=dm.screen().root,
                    window=window,
                    same_screen=0,
                    child=Xlib.X.NONE,
                    root_x=0,
                    root_y=0,
                    event_x=0,
                    event_y=0,
                )
            )

    def _resolve_dead(self, key):
        """Tries to resolve a dead key.

        :param str identifier: The identifier to resolve.
        """
        # pylint: disable=W0702; we want to ignore errors
        try:
            keysym, _ = SYMBOLS[CHARS[key.combining]]
        except Exception:
            return None
        # pylint: enable=W0702

        if keysym not in self.keyboard_mapping:
            return None

        return keysym

    def _resolve_special(self, key):
        """Tries to resolve a special key.

        A special key has the :attr:`~KeyCode.vk` attribute set.

        :param KeyCode key: The key to resolve.
        """
        if not key.vk:
            return None

        return key.vk

    def _resolve_normal(self, key):
        """Tries to resolve a normal key.

        A normal key exists on the keyboard, and is typed by pressing
        and releasing a simple key, possibly in combination with a modifier.

        :param KeyCode key: The key to resolve.
        """
        keysym = self._key_to_keysym(key)
        if keysym is None:
            return None

        if keysym not in self.keyboard_mapping:
            return None

        return keysym

    def _resolve_borrowed(self, key):
        """Tries to resolve a key by looking up the already borrowed *keysyms*.

        A borrowed *keysym* does not exist on the keyboard, but has been
        temporarily added to the layout.

        :param KeyCode key: The key to resolve.
        """
        keysym = self._key_to_keysym(key)
        if keysym is None:
            return None

        with self._borrow_lock:
            if keysym not in self._borrows:
                return None

        return keysym

    def _resolve_borrowing(self, key):
        """Tries to resolve a key by modifying the layout temporarily.

        A borrowed *keysym* does not exist on the keyboard, but is temporarily
        added to the layout.

        :param KeyCode key: The key to resolve.
        """
        keysym = self._key_to_keysym(key)
        if keysym is None:
            return None

        mapping = self._display.get_keyboard_mapping(8, 255 - 8)

        def i2kc(index):
            return index + 8

        def kc2i(keycode):
            return keycode - 8

        #: Finds a keycode and index by looking at already used keycodes
        def reuse():
            for _, (keycode, _, _) in self._borrows.items():
                keycodes = mapping[kc2i(keycode)]

                # Only the first four items are addressable by X
                for index in range(4):
                    if not keycodes[index]:
                        return keycode, index

        #: Finds a keycode and index by using a new keycode
        def borrow():
            for i, keycodes in enumerate(mapping):
                if not any(keycodes):
                    return i2kc(i), 0

        #: Finds a keycode and index by reusing an old, unused one
        def overwrite():
            for keysym, (keycode, index, count) in self._borrows.items():
                if count < 1:
                    del self._borrows[keysym]
                    return keycode, index

        #: Registers a keycode for a specific key and modifier state
        def register(dm, keycode, index):
            i = kc2i(keycode)
            mapping[i][index] = keysym
            dm.change_keyboard_mapping(keycode, mapping[i : i + 1])
            self._borrows[keysym] = (keycode, index, 0)

        try:
            with display_manager(self._display) as dm, self._borrow_lock as _:
                # First try an already used keycode, then try a new one, and
                # fall back on reusing one that is not currently pressed
                register(dm, *(reuse() or borrow() or overwrite()))
            return keysym

        except TypeError:
            return None

    def _key_to_keysym(self, key):
        """Converts a character key code to a *keysym*.

        :param KeyCode key: The key code.

        :return: a keysym if found
        :rtype: int or None
        """
        symbol = CHARS.get(key.char, None)
        if symbol is None:
            return None

        # pylint: disable=W0702; we want to ignore errors
        try:
            return symbol_to_keysym(symbol)
        except Exception:
            try:
                return SYMBOLS[symbol][0]
            except Exception:
                return None
        # pylint: enable=W0702

    def _shift_mask(self, modifiers):
        """The *X* modifier mask to apply for a set of modifiers.

        :param set modifiers: A set of active modifiers for which to get the
            shift mask.
        """
        return (
            0
            | (self.ALT_MASK if Key.alt in modifiers else 0)
            | (self.ALT_GR_MASK if Key.alt_gr in modifiers else 0)
            | (self.CTRL_MASK if Key.ctrl in modifiers else 0)
            | (self.SHIFT_MASK if Key.shift in modifiers else 0)
        )

    def _update_keyboard_mapping(self):
        """Updates the keyboard mapping."""
        with display_manager(self._display) as dm:
            self._keyboard_mapping = keyboard_mapping(dm)


@Controller._receiver
class Listener(ListenerMixin, _base.Listener):
    _EVENTS = (Xlib.X.KeyPress, Xlib.X.KeyRelease)

    #: A mapping from keysym to special key
    _SPECIAL_KEYS = {key.value.vk: key for key in Key}

    #: A mapping from numeric keypad keys to keys
    _KEYPAD_KEYS = {
        KEYPAD_KEYS["KP_0"]: KeyCode.from_char("0"),
        KEYPAD_KEYS["KP_1"]: KeyCode.from_char("1"),
        KEYPAD_KEYS["KP_2"]: KeyCode.from_char("2"),
        KEYPAD_KEYS["KP_3"]: KeyCode.from_char("3"),
        KEYPAD_KEYS["KP_4"]: KeyCode.from_char("4"),
        KEYPAD_KEYS["KP_5"]: KeyCode.from_char("5"),
        KEYPAD_KEYS["KP_6"]: KeyCode.from_char("6"),
        KEYPAD_KEYS["KP_7"]: KeyCode.from_char("7"),
        KEYPAD_KEYS["KP_8"]: KeyCode.from_char("8"),
        KEYPAD_KEYS["KP_9"]: KeyCode.from_char("9"),
        KEYPAD_KEYS["KP_Add"]: KeyCode.from_char("+"),
        KEYPAD_KEYS["KP_Decimal"]: KeyCode.from_char(","),
        KEYPAD_KEYS["KP_Delete"]: Key.delete,
        KEYPAD_KEYS["KP_Divide"]: KeyCode.from_char("/"),
        KEYPAD_KEYS["KP_Down"]: Key.down,
        KEYPAD_KEYS["KP_End"]: Key.end,
        KEYPAD_KEYS["KP_Enter"]: Key.enter,
        KEYPAD_KEYS["KP_Equal"]: KeyCode.from_char("="),
        KEYPAD_KEYS["KP_F1"]: Key.f1,
        KEYPAD_KEYS["KP_F2"]: Key.f2,
        KEYPAD_KEYS["KP_F3"]: Key.f3,
        KEYPAD_KEYS["KP_F4"]: Key.f4,
        KEYPAD_KEYS["KP_Home"]: Key.home,
        KEYPAD_KEYS["KP_Insert"]: Key.insert,
        KEYPAD_KEYS["KP_Left"]: Key.left,
        KEYPAD_KEYS["KP_Multiply"]: KeyCode.from_char("*"),
        KEYPAD_KEYS["KP_Page_Down"]: Key.page_down,
        KEYPAD_KEYS["KP_Page_Up"]: Key.page_up,
        KEYPAD_KEYS["KP_Right"]: Key.right,
        KEYPAD_KEYS["KP_Space"]: Key.space,
        KEYPAD_KEYS["KP_Subtract"]: KeyCode.from_char("-"),
        KEYPAD_KEYS["KP_Tab"]: Key.tab,
        KEYPAD_KEYS["KP_Up"]: Key.up,
    }

    def __init__(self, *args, **kwargs):
        super(Listener, self).__init__(*args, **kwargs)
        self._keyboard_mapping = None

    def _run(self):
        with self._receive():
            super(Listener, self)._run()

    def _initialize(self, display):
        # Get the keyboard mapping to be able to translate events details to
        # key codes
        min_keycode = display.display.info.min_keycode
        keycode_count = display.display.info.max_keycode - min_keycode + 1
        self._keyboard_mapping = display.get_keyboard_mapping(
            min_keycode, keycode_count
        )

    def _handle(self, display, event):
        # Convert the event to a KeyCode; this may fail, and in that case we
        # pass None
        try:
            key = self._event_to_key(display, event)
        except IndexError:
            key = None

        if event.type == Xlib.X.KeyPress:
            self.on_press(key)

        elif event.type == Xlib.X.KeyRelease:
            self.on_release(key)

    def _suppress_start(self, display):
        display.screen().root.grab_keyboard(
            self._event_mask,
            Xlib.X.GrabModeAsync,
            Xlib.X.GrabModeAsync,
            Xlib.X.CurrentTime,
        )

    def _suppress_stop(self, display):
        display.ungrab_keyboard(Xlib.X.CurrentTime)

    def _on_fake_event(self, key, is_press):
        """The handler for fake press events sent by the controllers.

        :param KeyCode key: The key pressed.

        :param bool is_press: Whether this is a press event.
        """
        (self.on_press if is_press else self.on_release)(
            self._SPECIAL_KEYS.get(key.vk, key)
        )

    def _keycode_to_keysym(self, display, keycode, index):
        """Converts a keycode and shift state index to a keysym.

        This method uses a simplified version of the *X* convention to locate
        the correct keysym in the display table: since this method is only used
        to locate special keys, alphanumeric keys are not treated specially.

        :param display: The current *X* display.

        :param keycode: The keycode.

        :param index: The shift state index.

        :return: a keysym
        """
        keysym = display.keycode_to_keysym(keycode, index)
        if keysym:
            return keysym
        elif index & 0x2:
            return self._keycode_to_keysym(display, keycode, index & ~0x2)
        elif index & 0x1:
            return self._keycode_to_keysym(display, keycode, index & ~0x1)
        else:
            return 0

    def _event_to_key(self, display, event):
        """Converts an *X* event to a :class:`KeyCode`.

        :param display: The current *X* display.

        :param event: The event to convert.

        :return: a :class:`pynput.keyboard.KeyCode`

        :raises IndexError: if the key code is invalid
        """
        keycode = event.detail
        index = shift_to_index(display, event.state)

        # First try special keys...
        keysym = self._keycode_to_keysym(display, keycode, index)
        if keysym in self._SPECIAL_KEYS:
            return self._SPECIAL_KEYS[keysym]
        elif keysym in self._KEYPAD_KEYS:
            # We must recalculate the index if numlock is active; index 1 is the
            # one to use
            try:
                return self._KEYPAD_KEYS[
                    self._keycode_to_keysym(
                        display, keycode, bool(event.state & numlock_mask(display))
                    )
                ]
            except KeyError:
                # Since we recalculated the key, this may happen
                pass

        # ...then try characters...
        name = KEYSYMS.get(keysym, None)
        if name is not None and name in SYMBOLS:
            char = SYMBOLS[name][1].upper() if index & 1 else SYMBOLS[name][1]
            if char in DEAD_KEYS:
                return KeyCode.from_dead(DEAD_KEYS[char], vk=keysym)
            else:
                return KeyCode.from_char(char, vk=keysym)

        # ...and fall back on a virtual key code
        return KeyCode.from_vk(keysym)



from numpy import array

from pitop.core import ImageFunctions
from pitop.core.data_structures import DotDict
from pitop.processing.core.vision_functions import (
    center_reposition,
    color_mask,
    find_centroid,
    find_largest_contour,
    get_object_target_lock_control_angle,
    import_opencv,
)


def calculate_blue_limits():
    # define range of blue color in HSV-> H: 0-179, S: 0-255, V: 0-255
    # broken out like this for easy conversion into CV units (will move somewhere more logical in future)
    hue_lower = 160
    hue_upper = 280
    sat_lower = 0.3
    sat_upper = 1.0
    val_lower = 0.5
    val_upper = 1.0
    cv_hue_lower = int(hue_lower / 2)
    cv_hue_upper = int(hue_upper / 2)
    cv_sat_lower = int(sat_lower * 255)
    cv_sat_upper = int(sat_upper * 255)
    cv_val_lower = int(val_lower * 255)
    cv_val_upper = int(val_upper * 255)
    lower_blue = array([cv_hue_lower, cv_sat_lower, cv_val_lower])
    upper_blue = array([cv_hue_upper, cv_sat_upper, cv_val_upper])
    return lower_blue, upper_blue


def process_frame_for_line(frame, image_format="PIL", process_image_width=320):
    cv2 = import_opencv()
    cv_frame = ImageFunctions.convert(frame, format="OpenCV")

    from imutils import resize

    resized_frame = resize(cv_frame, width=process_image_width)
    hsv_lower, hsv_upper = calculate_blue_limits()
    image_mask = color_mask(resized_frame, hsv_lower, hsv_upper)
    line_contour = find_largest_contour(image_mask)

    centroid = None
    scaled_image_centroid = None
    rectangle_dimensions = None
    angle = None
    if line_contour is not None:
        # find centroid of contour
        scaled_image_centroid = find_centroid(line_contour)
        centroid = center_reposition(scaled_image_centroid, resized_frame)
        bounding_rectangle = cv2.boundingRect(line_contour)
        rectangle_dimensions = bounding_rectangle[2:5]
        angle = get_object_target_lock_control_angle(centroid, resized_frame)

    robot_view_img = robot_view(
        resized_frame, image_mask, line_contour, scaled_image_centroid
    )

    if image_format.lower() != "opencv":
        robot_view_img = ImageFunctions.convert(robot_view_img, format="PIL")

    return DotDict(
        {
            "line_center": centroid,
            "robot_view": robot_view_img,
            "rectangle_dimensions": rectangle_dimensions,
            "angle": angle,
        }
    )


def robot_view(frame, image_mask, line_contour, centroid):
    cv2 = import_opencv()
    masked_image = cv2.bitwise_and(frame, frame, mask=image_mask)

    # draw contour lines on robot view
    if line_contour is not None:
        cv2.drawContours(masked_image, [line_contour], 0, (100, 60, 240), 2)
        draw_contour_bound(masked_image, line_contour)

    if centroid:
        cv2.drawMarker(
            masked_image,
            (centroid[0], centroid[1]),
            (100, 60, 240),
            markerType=cv2.MARKER_CROSS,
            markerSize=20,
            thickness=4,
            line_type=cv2.FILLED,
        )
    return masked_image


def draw_contour_bound(image, contour):
    cv2 = import_opencv()
    x, y, w, h = cv2.boundingRect(contour)
    cv2.rectangle(image, (x, y), (x + w, y + h), (0, 255, 0), 2)



import os
import pickle

# directory where calibration output pickle file is located
calibration_outputs_dir = ""
script_dir = os.path.dirname(os.path.realpath(__file__))
abs_file_path = os.path.join(script_dir, calibration_outputs_dir)

# Filename used to save the camera calibration result (mtx, dist)
calibration_mtx_dist_filename = "camera_cal_dist_pickle_640-480.p"

# Camera resolution used for calibration
calibration_width = 640
calibration_height = 480


def load_camera_cal(width: int, height: int):
    """Read in the saved camera matrix and distortion coefficients These are
    the arrays we calculated using cv2.calibrateCamera() Also scales
    calibration matrix based on resolution being used."""

    dist_pickle = pickle.load(
        open(os.path.join(abs_file_path, calibration_mtx_dist_filename), "rb")
    )

    mtx = dist_pickle["mtx"]
    dist = dist_pickle["dist"]

    scale_factor_x = width / calibration_width
    scale_factor_y = height / calibration_height

    mtx[0][0] = mtx[0][0] * scale_factor_x
    mtx[1][1] = mtx[1][1] * scale_factor_y
    mtx[0][2] = mtx[0][2] * scale_factor_x
    mtx[1][2] = mtx[1][2] * scale_factor_y

    return mtx, dist



from abc import ABC, abstractmethod
from os import mkdir, path
from pathlib import Path
from time import strftime


class CaptureActionBase(ABC):
    """Abstract class from which all capture actions classes must inherit
    from."""

    @abstractmethod
    def process(self, frame):
        pass

    @abstractmethod
    def stop(self):
        pass

    def _get_output_filename(self, directory, extension):
        return path.join(
            directory, "output_" + strftime("%Y-%m-%d-%H-%M-%S") + "." + extension
        )

    def _create_output_directory(self):
        output_directory = path.join(str(Path.home()), "Camera")

        if not path.isdir(output_directory):
            mkdir(output_directory)

        return output_directory



from .base import BaseBlueprint
from .controller import ControllerBlueprint
from .messaging import MessagingBlueprint
from .rover import RoverControllerBlueprint
from .video import VideoBlueprint
from .webcomponents import WebComponentsBlueprint



class Error(Exception):
    """Base class for exceptions in this library."""

    pass


class DeviceNotFoundError(Error):
    """Exception raised when a device cannot be found."""


class DevicePermissionError(Error):
    """Exception raised when permission to access the device is denied."""


class DeviceAddressError(Error):
    """Exception raised when an invalid device address is detected."""


class DeviceDisplayModeError(Error):
    """Exception raised when an invalid device display mode is detected."""


class UnsupportedPlatform(Error):
    """Exception raised when trying to use the library on an incompatible
    system."""



from os import path
from subprocess import getstatusoutput
from sys import stderr

from pitop.common.common_ids import FirmwareDeviceID
from pitop.common.common_names import DeviceName
from pitop.common.firmware_device import FirmwareDevice

from .cli_base import CliBaseClass, PitopCliInvalidArgument
from .imu_calibration import ImuCalibration


class ImuCLI(CliBaseClass):
    parser_help = "Expansion Plate IMU utilities"
    cli_name = "imu"

    def __init__(self, args) -> None:
        self.args = args

    def run(self) -> int:
        # Check if device is a pi-top[4]
        from pitop.system import device_type

        is_pi_four = device_type() == DeviceName.pi_top_4.value
        if not is_pi_four:
            print("This CLI only runs on a pi-top [4].", file=stderr)
            return 1

        # Check if Expansion Plate is connected
        expansion_plate_info = FirmwareDevice.device_info.get(
            FirmwareDeviceID.pt4_expansion_plate
        )
        i2c_address = expansion_plate_info.get("i2c_addr")
        try:
            expansion_plate_connected = (
                getstatusoutput(f"pt-i2cdetect {i2c_address}")[0] == 0
            )
        except Exception:
            expansion_plate_connected = False
        if not expansion_plate_connected:
            print("No Expansion Plate detected - unable to calibrate IMU.", file=stderr)
            return 1

        if self.args.imu_subcommand == "calibrate":
            if self.args.path and not path.isdir(self.args.path):
                print(f"{self.args.path} is not a valid directory")
                return 1

            imu_cal = ImuCalibration()
            imu_cal.calibrate_magnetometer()
            imu_cal.plot_graphs(self.args.path)
        else:
            raise PitopCliInvalidArgument("oops")
        return 0

    @classmethod
    def add_parser_arguments(cls, parser) -> None:

        subparser = parser.add_subparsers(
            title="IMU utility", description=cls.parser_help, dest="imu_subcommand"
        )

        calibrate_parser = subparser.add_parser(
            "calibrate", help="Calibrate the magnetometer"
        )
        calibrate_parser.add_argument(
            "-p",
            "--path",
            type=str,
            help="Directory for storing calibration graph data",
        )



from enum import Enum
from inspect import signature
from threading import Event, Thread

from pitop.core import ImageFunctions
from pitop.core.mixins import Recreatable, Stateful
from pitop.pma.common import type_check

from .core import CameraTypes, FileSystemCamera, FrameHandler, UsbCamera
from .core.capture_actions import CaptureActions


class Camera(Stateful, Recreatable):
    """Provides a variety of high-level functionality for using the PMA USB
    Camera, including capturing images and video, and processing image data
    from the camera.

    :type index: int
    :param index:
        ID of the video capturing device to open.
        Passing `None` will cause the backend to autodetect the
        available video capture devices and attempt to use them.
    """

    __VALID_FORMATS = ("opencv", "pil")

    def __init__(
        self,
        index=None,
        resolution=(640, 480),
        camera_type=CameraTypes.USB_CAMERA,
        path_to_images="",
        format="PIL",
        flip_top_bottom: bool = False,
        flip_left_right: bool = False,
        rotate_angle=0,
        name="camera",
    ):
        # Initialise private variables
        self._resolution = resolution
        self._format = None

        # Set format using setter property function
        self.format = format

        # Frame callback
        self.on_frame = None

        # Internal
        self._index = index
        self._camera_type = CameraTypes(camera_type)
        self._path_to_images = path_to_images
        self._rotate_angle = rotate_angle
        self._flip_top_bottom = flip_top_bottom
        self._flip_left_right = flip_left_right

        if self._camera_type == CameraTypes.USB_CAMERA:
            self.__camera = UsbCamera(
                index=self._index,
                resolution=self._resolution,
                flip_top_bottom=flip_top_bottom,
                flip_left_right=flip_left_right,
                rotate_angle=self._rotate_angle,
            )

        elif self._camera_type == CameraTypes.FILE_SYSTEM_CAMERA:
            self.__camera = FileSystemCamera(self._path_to_images)

        self.__continue_processing = True
        self.__frame_handler = FrameHandler()
        self.__new_frame_event = Event()
        self.__process_image_thread = Thread(
            target=self.__process_camera_output, daemon=True
        )
        self.__process_image_thread.start()

        self.name = name
        Stateful.__init__(self)
        Recreatable.__init__(
            self,
            config_dict={
                "index": index,
                "resolution": resolution,
                "camera_type": camera_type.value
                if isinstance(camera_type, Enum)
                else camera_type,
                "path_to_images": path_to_images,
                "format": format,
                "name": self.name,
                "flip_top_bottom": self._flip_top_bottom,
                "flip_left_right": self._flip_left_right,
                "rotate_angle": self._rotate_angle,
            },
        )

    @property
    def own_state(self):
        return {
            "running": lambda: self.__process_image_thread.is_alive(),
            "capture_actions": lambda: self.__frame_handler.current_actions(),
        }

    @classmethod
    def from_file_system(cls, path_to_images: str):
        """Alternative classmethod to create an instance of a :class:`Camera`
        object using a :data:`FileSystemCamera`"""
        return cls(
            camera_type=CameraTypes.FILE_SYSTEM_CAMERA, path_to_images=path_to_images
        )

    @property
    def format(self):
        return self._format

    @format.setter
    def format(self, format_value):
        ImageFunctions.image_format_check(format_value)
        self._format = format_value.lower()

    @classmethod
    def from_usb(cls, index=None):
        """Alternative classmethod to create an instance of a :class:`Camera`
        object using a :data:`UsbCamera`"""
        return cls(camera_type=CameraTypes.USB_CAMERA, index=index)

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.__camera = None
        self.__continue_processing = False
        if self.__process_image_thread.is_alive():
            self.__process_image_thread.join()

    def is_recording(self):
        """Returns True if recording mode is enabled."""

        return self.__frame_handler.is_running_action(
            CaptureActions.CAPTURE_VIDEO_TO_FILE
        )

    def is_detecting_motion(self):
        """Returns True if motion detection mode is enabled."""

        return self.__frame_handler.is_running_action(CaptureActions.DETECT_MOTION)

    @type_check
    def capture_image(self, output_file_name=""):
        """Capture a single frame image to file.

        .. note::
            If no :data:`output_file_name` argument is provided, images will be stored in `~/Camera`.

        :type output_file_name: str
        :param output_file_name:
            The filename into which to write the image.
        """

        self.__frame_handler.register_action(
            CaptureActions.CAPTURE_SINGLE_FRAME, {"output_file_name": output_file_name}
        )

    @type_check
    def start_video_capture(self, output_file_name="", fps=20.0, resolution=None):
        """Begin capturing video from the camera.

        .. note::
            If no :data:`output_file_name` argument is provided, video will be stored in `~/Camera`.

        :type output_file_name: str
        :param output_file_name:
            The filename into which to write the video.
        :type fps: int or float
        :param fps:
            The framerate to use for the captured video. Defaults to 20.0 fps
        :type resolution: tuple
        :param resolution:
            The resolution to use for the captured video. Defaults to (640, 368)
        """

        args = {
            "output_file_name": output_file_name,
            "fps": fps,
            "resolution": resolution,
        }
        self.__frame_handler.register_action(CaptureActions.CAPTURE_VIDEO_TO_FILE, args)

    def stop_video_capture(self):
        """Stop capturing video from the camera.

        Does nothing unless :class:`start_video_capture` has been
        called.
        """

        self.__frame_handler.remove_action(CaptureActions.CAPTURE_VIDEO_TO_FILE)

    @type_check
    def start_detecting_motion(
        self, callback_on_motion, moving_object_minimum_area=300
    ):
        """Begin processing image data from the camera, attempting to detect
        motion. When motion is detected, call the function passed in.

        .. warning::
            The callback function can take either no arguments or only one, which will be used to provide the image back
            to the user when motion is detected.
            If a callback with another signature is received, the method will raise an exception.

        :type callback_on_motion: function
        :param callback_on_motion:
            A callback function that will be called when motion is detected.
        :type moving_object_minimum_area: int
        :param moving_object_minimum_area:
            The sensitivity of the motion detection, measured as the area of pixels changing between frames that constitutes motion.
        """

        args = {
            "callback_on_motion": callback_on_motion,
            "moving_object_minimum_area": moving_object_minimum_area,
        }
        callback_signature = signature(callback_on_motion)
        if len(callback_signature.parameters) > 1:
            raise ValueError(
                "Invalid callback signature: it should receive at most one argument."
            )
        self.__frame_handler.register_action(CaptureActions.DETECT_MOTION, args)

    def stop_detecting_motion(self):
        """Stop running the motion detection processing.

        Does nothing unless :class:`start_detecting_motion` has been
        called.
        """

        self.__frame_handler.remove_action(CaptureActions.DETECT_MOTION)

    @type_check
    def start_handling_frames(self, callback_on_frame, frame_interval=1, format=None):
        """Begin calling the passed callback with each new frame, allowing for
        custom processing.

        .. warning::
            The callback function can take either no arguments or only one, which will be used to provide the image back
            to the user.
            If a callback with another signature is received, the method will raise an exception.

        :type callback_on_frame: function
        :param callback_on_frame:
            A callback function that will be called every :class:`frame_interval` camera frames.
        :type frame_interval: int
        :param frame_interval:
            The callback will run every frame_interval frames, decreasing the frame rate of processing. Defaults to 1.

        :type format: string
        :param format:
            DEPRECATED. Set 'camera.format' directly, and call this function directly instead.
        """
        if format is not None:
            print(
                "'format' is no longer supported in this function. "
                "Please set the 'camera.format' property directly, and call this function without 'format' parameter."
            )
        args = {
            "callback_on_frame": callback_on_frame,
            "frame_interval": frame_interval,
            "format": self.format,
        }
        callback_signature = signature(callback_on_frame)
        if len(callback_signature.parameters) == 0:
            raise ValueError(
                "Invalid callback signature: it should receive at least one argument."
            )
        self.__frame_handler.register_action(CaptureActions.HANDLE_FRAME, args)

    def stop_handling_frames(self):
        """Stops handling camera frames.

        Does nothing unless :class:`start_handling_frames` has been
        called.
        """

        self.__frame_handler.remove_action(CaptureActions.HANDLE_FRAME)

    def __get_processed_current_frame(self):
        image = self.__frame_handler.frame

        if self.format.lower() == "opencv":
            image = ImageFunctions.convert(image, format="opencv")

        return image

    def __process_camera_output(self):
        while self.__camera and self.__continue_processing is True:
            self.__frame_handler.frame = self.__camera.get_frame()
            self.__new_frame_event.set()

            if callable(self.on_frame):
                self.on_frame(self.__get_processed_current_frame())

            try:
                self.__frame_handler.process()
            except Exception as e:
                print(f"Error in camera frame handler: {e}")

    def current_frame(self, format=None):
        """Returns the latest frame captured by the camera. This method is non-
        blocking and can return the same frame multiple times.

        By default the returned image is formatted as a :class:`PIL.Image.Image`.

        :type format: string
        :param format:
            DEPRECATED. Set 'camera.format' directly, and call this function directly instead.
        """
        if format is not None:
            print(
                "'format' is no longer supported in this function. "
                "Please set the 'camera.format' property directly, and call this function without 'format' parameter."
            )

        return self.__get_processed_current_frame()

    def get_frame(self, format=None):
        """Returns the next frame captured by the camera. This method blocks
        until a new frame is available.

        :type format: string
        :param format:
            DEPRECATED. Set 'camera.format' directly, and call this function directly instead.
        """
        if format is not None:
            print(
                "'format' is no longer supported in this function. "
                "Please set the 'camera.format' property directly, and call this function without 'format' parameter."
            )

        self.__new_frame_event.wait()
        self.__new_frame_event.clear()

        return self.__get_processed_current_frame()



from flask import Flask, render_template, send_from_directory
from flask_cors import CORS
from flask_sockets import Sockets
from gevent.pywsgi import WSGIServer
from geventwebsocket.handler import WebSocketHandler

from pitop.common.formatting import is_url
from pitop.common.sys_info import get_internal_ip

from .blueprints.base import BaseBlueprint


def create_app(
    import_name="__main__", *args, template_folder=".", static_folder=".", **kargs
):
    app = Flask(
        import_name,
        *args,
        template_folder=template_folder,
        static_folder=static_folder,
        **kargs,
    )

    CORS(app)

    @app.route("/status")
    def status():
        return "OK"

    @app.route("/", defaults={"path": "index"})
    @app.route("/<path>")
    def send_file(path):
        try:
            return render_template(f"{path}.html")
        except Exception:
            return send_from_directory(app.static_folder, path)

    @app.after_request
    def add_header(req):
        req.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
        req.headers["Pragma"] = "no-cache"
        req.headers["Expires"] = "0"
        return req

    app.config["TEMPLATES_AUTO_RELOAD"] = True

    return app


class WebServer(WSGIServer):
    def __init__(self, port=8070, app=create_app(), blueprints=[BaseBlueprint()]):
        self.port = port
        self.app = app
        self.sockets = Sockets(app)

        with self.app.app_context():
            for blueprint in blueprints:
                self.app.register_blueprint(blueprint, sockets=self.sockets)

        WSGIServer.__init__(
            self, ("0.0.0.0", port), self.app, handler_class=WebSocketHandler
        )

    def _log_address(self):
        ip_addresses = list()
        for interface in ("wlan0", "ptusb0", "lo", "en0"):
            ip_address = get_internal_ip(interface)
            if is_url("http://" + ip_address):
                ip_addresses.append(ip_address)

        print("WebServer is listening at:")
        if len(ip_addresses) > 0:
            for ip_address in ip_addresses:
                print(f"\t- http://{ip_address}:{self.port}/")
        else:
            print(f"\t- http://localhost:{self.port}/ (on same device)")

    def start(self):
        self._log_address()
        super().start()
